	page	59,132
	title	UIDE -- DOS "Universal IDE" Caching Driver.
;
; This is a general-purpose caching driver for DOS drives addressed by
; Int 13h BIOS I-O requests.   UIDE runs SATA and UltraDMA disks using
; internal logic.   It caches up to 32 BIOS units, including A: and B:
; diskettes (if present) through diskette BIOS logic.   UIDE also runs
; up to 4 CD/DVD drives and caches their file-input requests.   CD/DVD
; "raw" input (audio, trackwriters) is supported but not cached.   I-O
; for up to 220 other "external" drives (USB, Firewire, etc.) can also
; be cached.   "Write Through" caching is done for all output.    UIDE
; uses XMS memory and caches from 5-MB to 1-Gigabyte of data!
;
; General Program Equations.
;
	.386p			;Allow use of 80386 instructions.
	.lfcond			;List conditionally-unassembled lines.
s	equ	<short>		;Default conditional jumps to "short".
STACK	equ	560		;Driver local-stack size.
RMAXLBA	equ	00006DD39h	;Redbook (audio) maximum LBA value.
COOKSL	equ	2048		;CD/DVD "cooked" sector length.
RAWSL	equ	2352		;CD/DVD "raw" sector length.
RDYTO	equ	008h		;Disk 389-msec minimum I-O timeout.
CMDTO	equ	00Ah		;CD/DVD 500-msec min. command timeout.
SEEKTO	equ	037h		;CD/DVD 3-second min. "seek"  timeout.
STARTTO	equ	07Fh		;CD/DVD 7-second min. startup timeout.
HDWRFL	equ	00410h		;BIOS hardware-installed flag.
DKTSTAT	equ	00441h		;BIOS diskette status byte.
BIOSTMR equ	0046Ch		;BIOS "tick" timer address.
HDISKS	equ	00475h		;BIOS hard-disk count address.
VDSFLAG equ	0047Bh		;BIOS "Virtual DMA" flag address.
HDI_OFS	equ	0048Eh-BIOSTMR	;BIOS hard-disk int. flag "offset".
MCHDWFL	equ	0048Fh		;Diskette hardware media-change flags.
IXM	equ	4096		;IOCTL transfer-length multiplier.
CR	equ	00Dh		;ASCII carriage-return.
LF	equ	00Ah		;ASCII line-feed.
TAB	equ	009h		;ASCII "tab".
;
; Driver-Size and Cache-Size Definitions.
;
MAXBIOS	equ	32		;Maximum BIOS units handled.
TBLCT	equ	640		;Starting cache-table entry count.
CDTYPE	equ	020h		;CD/DVD device type.
EXTYPE	equ	030h		;+010h for each new "internal driver".
;
; Driver Error Return Codes.
;
MCHANGE	equ	006h		;Diskette "media change".
GENERR	equ	00Ch		;CD/DVD "general error".
DMAERR	equ	00Fh		;Hard disk DMA error.
CTLRERR equ	020h		;Hard disk controller not-ready.
DISKERR equ	0AAh		;Hard disk drive not-ready.
FAULTED	equ	0CCh		;Hard disk drive FAULTED.
HARDERR equ	0E0h		;Hard disk I-O error.
XMSERR	equ	0FFh		;XMS memory error.
;
; "Legacy IDE" Controller I-O Base Addresses.
;
NPDATA	equ	001F0h		;Normal primary      base address.
NSDATA	equ	00170h		;Normal secondary    base address.
APDATA	equ	001E8h		;Alternate primary   base address.
ASDATA	equ	00168h		;Alternate secondary base address.
;
; IDE Controller Register Offsets.
;
CDATA	equ	0		;Data port offset.
CDSEL	equ	6		;Disk-select and upper LBA offset.
CCMD	equ	7		;Command-register offset.
CSTAT	equ	7		;Primary-status register offset.
;
; Controller Status and Command Definitions.
;
BSY	equ	080h		;IDE controller is busy.
RDY	equ	040h		;IDE disk is "ready".
FLT	equ	020h		;IDE disk has a "fault".
DRQ	equ	008h		;IDE data request.
ERR	equ	001h		;IDE general error flag.
DMI	equ	004h		;DMA interrupt occured.
DME	equ	002h		;DMA error occurred.
MSEL	equ	0A0h		;"Master" device-select bits.
SSEL	equ	0B0h		;"Slave"  device-select bits.
DRCMD	equ	0C8h		;DMA read command (write is 0CAh,
				;    LBA48 commands are 025h/035h).
LBABITS equ	0E0h		;Fixed LBA command bits.
;
; Byte, Word, and Double-Word Definitions.
;
BDF	struc
lb	db	?
hb	db	?
BDF	ends

WDF	struc
lw	dw	?
hw	dw	?
WDF	ends

DDF	struc
dwd	dd	?
DDF	ends
;
; LBA "Device Address Packet" Layout.
;
DAP	struc
DapPL	db	?		;Packet length.
	db	?		;(Reserved).
DapSC	db	?		;I-O sector count.
	db	?		;(Reserved).
DapBuf	dd	?		;I-O buffer address (segment:offset).
DapLBA	dw	?		;48-bit logical block address (LBA).
DapLBA1	dd	?
DAP	ends
;
; DOS "Request Packet" Layout.
;
RP	struc
RPHLen	db	?		;Header byte count.
RPSubU	db	?		;Subunit number.
RPOp	db	?		;Opcode.
RPStat	dw	?		;Status word.
	db	8 dup (?)	;(Unused by us).
RPUnit	db	?		;Number of units found.
RPLen	dw	?		;Resident driver 32-bit length.
RPSeg	dw	?		;(Segment must ALWAYS be set!).
RPCL	dd	?		;Command-line data pointer.
RP	ends
RPERR	equ	08003h		;Packet "error" flags.
RPDON	equ	00100h		;Packet "done" flag.
RPBUSY	equ	00200h		;Packet "busy" flag.
;
; IOCTL "Request Packet" Layout.
;
IOC	struc
	db	13 dup (?)	;Request "header" (unused by us).
	db	?		;Media descriptor byte (Unused by us).
IOCAdr	dd	? 		;Data-transfer address.
IOCLen	dw	?		;Data-transfer length.
	dw	?		;Starting sector (unused by us).
	dd	?		;Volume I.D. pointer (unused by us).
IOC	ends
;
; Read Long "Request Packet" Layout.
;
RL	struc
	db	13 dup (?)	;Request "header" (unused by us).
RLAM	db	?		;Addressing mode.
RLAddr	dd	?		;Data-transfer address.
RLSC	dw	?		;Data-transfer sector count.
RLSec	dd	?		;Starting sector number.
RLDM	db	?		;Data-transfer mode.
RLIntlv	db	?		;Interleave size.
RLISkip	db	?		;Interleave skip factor.
RL	ends
;
; Segment Declarations.
;
CODE	segment	public use16 'CODE'
	assume	cs:CODE,ds:CODE
;
; DOS Driver Device Header.
;
@	dd	0FFFFFFFFh	;Link to next header block.
	dw	0C800h		;Driver "device attributes".
	dw	(Strat-@)	;"Strategy" routine pointer.
DevIntP	dw	(I_Init-@)	;Device-Interrupt routine pointer.
DvrNam	db	'UDVD1   '	;Driver name for use by SHCDX33C, etc.
	dw	0		;(Reserved).
	db	0		;First assigned drive letter.
CDUnits	db	0		;Number of CD/DVD drives (1 thru 3).
ExEntrP	dw	(ExEntr-@)	;"External Entry" pointer, MUST be at
				;  address 00016h!  Other drivers can
				;  then access "ExEntr" etc. by using
                                ;  our code-segment and this pointer.
;
; VDS Parameter Block.
;
VDSLn	dd	(NormSz1-@)	;VDS block length.
VDSOf	dd	0		;VDS 32-bit buffer offset.
VDSSg	dd	0		;VDS 16-bit segment (hi-order zero).
VDSAd	dd	0		;VDS 32-bit physical buffer address.
;
; UltraDMA Command-List.
;
IOAdr	dd	0		;DMA 32-bit buffer address.
IOLen	dd	080000000h	;DMA byte count and "end" flag.
;
; ATAPI "Packet" Area (always 12 bytes for a CD/DVD).
;
Packet	db	0		;Opcode.
	db	0		;Unused (LUN and reserved).
PktLBA	dd	0		;CD/DVD logical block address.
PktLH	db	0		;"Transfer length" (sector count).
PktLn	dw	0		;Middle- and low-order sector count.
PktRM	db	0		;Read mode ("Raw" Read Long only).
	dw	0		;Unused ATAPI "pad" bytes (required).
;
; "PRD" Address and IDE Parameter Area.
;
PRDAd	dd	0		;PRD 32-bit command addr. (Init set).
	db	0		;IDE "upper" sector count (always 0).
LBA2	db	0,0,0		;IDE "upper" LBA bits 24-47.
SecCt	db	0		;IDE "lower" sector count.
LBA	db	0,0,0		;IDE "lower" LBA bits 0-23.
DSCmd	db	0		;IDE disk-select and LBA commands.
IOCmd	db	0		;IDE command byte.
Try	equ	LBA2.lb		;CD/DVD I-O retry counter.
UserB	equ	SecCt.dwd	;CD/DVD working I-O buffer address.
UserL	equ	DSCmd.lw	;CD/DVD working I-O buffer length.
;
; General Driver "Data Page" Variables.   Note that most of these
;   overlay the driver "Final Initialization" routine below.
;
IOF	db	001h		;I-O busy flag (80h busy, 01h flush).
VLF	db	0		;VDS lock flag (01h = buffer locked).
DMAAd	dw	0		;DMA status/command register address.
IdeDA	dw	0		;IDE data register address.
RqPkt	dd	0		;Request I-O packet address.
RqBuf	dd	0		;Request I-O buffer address.
RqXMS	dd	128		;Request XMS buffer offset.
RqSec	db	0		;Request I-O sector count.
RqCmd	equ	[$].lb		;Request I-O command bits.
RqCSC	equ	[$+1].lb	;Request current-block sector count.
RqTyp	equ	[$+2].lb	;Request device-type flag.
RqLBA	equ	[$+3].lw	;Request initial LBA address.
RqUNo	equ	[$+9].lb	;Request "cache unit" number.
CBLBA	equ	[$+11].lw	;Current-buffer initial LBA.
CBUNo	equ	[$+17].lb	;Current-buffer "cache unit" number.
CBSec	equ	[$+18].lb	;Current-buffer sector count.
CBLst	equ	[$+19].lw	;Current-buffer "last" LRU index.
CBNxt	equ	[$+21].lw	;Current-buffer "next" LRU index.
AudAP	equ	[$+23].lw	;CD/DVD audio-start address pointer.
XMF	equ	[$+25].lb	;CD/DVD XMS move  flag (01h if so).
DMF	equ	[$+26].lb	;CD/DVD DMA input flag (01h if so).
STLmt	equ	[$+27].lw	;Cache binary-search table limit.
LUTop	equ	[$+29].lw	;Least-recently-used "top" indexes.
LUEnd	equ	[$+31].lw	;Least-recently-used "end" indexes.
;
; Driver "Final Initialization" Routine.   The driver stack must be
;   cleared here (not above!), since this driver varies in size and
;   its final run-time stack may be set WITHIN the main init logic!
;
I_End:	cld			;Clear driver stack (helps debug).
	xor	ax,ax
	mov	cx,STACK
	mov	di,@Stack1
	sub	di,cx
	shr	cx,1
	rep	stosw
	mov	ax,02513h	;"Hook" this driver into Int 13h.
	mov	dx,offset (Entry-@)
	int	021h
I_Fail:	popad			;Reload all 32-bit CPU registers.
I_Exit:	pop	dx		;Reload 16-bit CPU regs. we used.
	pop	bx
	pop	ax
	pop	es		;Reload CPU segment registers.
	pop	ds
	popf			;Reload CPU flags and exit.
	retf
;
; BIOS "Active Units" Tables.
;
Units	db   MAXBIOS dup (0FFh)	;BIOS unit-number table, 0FFh unused.
TypeF	db   MAXBIOS dup (0)	;Device type and "sub-unit":
				;  000h-00Fh:  SATA/UltraDMA disk.
				;  010h-01Fh:  SATA/UltraDMA disk.
				;  020h:       CD and DVD drives.
				;  030h-03Fh:  Driver 2 devices.
				;  040h-04Fh:  Driver 3 devices.
				;  050h-05Fh:  Driver 4 devices.
				;  060h-06Fh:  Driver 5 devices.
				;  070h:       "External" devices.
				;  080h:       BIOS-driven hard disk.
				;  0C0h:       BIOS-driven diskette.
CHSec	db   MAXBIOS dup (0)	;Sectors-per-head table.
CHSHd	db   MAXBIOS dup (0)	;Heads-per-cylinder table.
;
; "Miscellaneous" Caching Variables.
;
LBABuf	dw	00082h,0,0	;"Calculated LBA" buffer.
;
; Controller I-O Address Table, set during Initialization.
;
Ctl1Tbl	dw	0FFFFh		;Controller 1 DMA base address.
Ctl1Pri	dw	NPDATA		;   "Legacy" primary   addresses.
Ctl1Sec	dw	NSDATA		;   "Legacy" secondary addresses.
Ctl2Tbl	dw	0FFFFh		;Controller 2 DMA base address.
Ctl2Pri	dw	APDATA		;   "Legacy" primary   addresses.
Ctl2Sec	dw	ASDATA		;   "Legacy" secondary addresses.
	dw	0FFFFh,0,0	;Controller 3 I-O addresses.
	dw	0FFFFh,0,0	;Controller 4 I-O addresses.
	dw	0FFFFh,0,0	;Controller 5 I-O addresses.
	dw	0FFFFh,0,0	;Controller 6 I-O addresses.
	dw	0FFFFh,0,0	;Controller 7 I-O addresses.
	dw	0FFFFh,0,0	;Controller 8 I-O addresses.
CtlTEnd	equ	$		   ;End of controller I-O tables.
CTLTSIZ	equ	(Ctl2Tbl-Ctl1Tbl)  ;Size of a controller I-O table.
;
; "Internal Driver" Handler Table.   Each points to a subroutine that
;   handles the "device types" shown below.   Along with "UdmaIO" and
;   "RLRead", users can add up to 4 more "internal" drivers, or up to
;   5 with no "external drives" (/DEXTUNITS omitted during assembly).
;
;   An "internal" driver works on the UIDE stack, can access all UIDE
;   variables, and must exit with a "retf" command.   It can post bit
;   7 of "RqCmd", indicating that an I-O request fit within one cache
;   block, and that block's XMS buffer was used to handle I-O.    The
;   main caching routine will then assume data has been "cached" into
;   XMS memory DURING the I-O request!    See the "UseCB" subroutine.
;
DvrTbl	dd	(UdmaIO-@)	;Type 00h - 0Fh:  SATA/UltraDMA disk.
Ud2Ptr	dd	(UdmaIO-@)	;Type 10h - 1Fh:  SATA/UltraDMA disk.
CDDPtr	dd	(RLRead-@)	;Type 20h:	  CD or DVD drive.
ExtDvr	equ	[$].dwd		;Type 30h+:       "External" drive.
;
; Cache "Working Buffer" (shares the CD/DVD Audio Function Buffer).
;
WBLBA	equ	[$+4].lw	;Working-buffer initial LBA.
WBUNo	equ	[$+10].lb	;Working-buffer "cache unit" number.
WBSec	equ	[$+11].lb	;Working-buffer sector count.
WBLst	equ	[$+12].lw	;Working-buffer "last" LRU index.
WBNxt	equ	[$+14].lw	;Working-buffer "next" LRU index.
;
; CD/DVD Audio Function Buffer (16 bytes) for most CD/DVD "audio"
;   requests.    Its variables below are only for initialization.
;
InBuf	equ	$
UTblP	dw	UTable		;Initialization unit table pointer.
NoDMA	db	0		;"No UltraDMA" flag.
UMode	dw	0		;UltraDMA "mode select" bits.
UFlag	db	0		;UltraDMA "mode valid" flags.
USNdx	db	5		;Unit-select index.
USelB	db	082h		;PCI "modes" and unit-select byte.
DVDTblA	dw	(DVD1Tbl-@)	;PCI "scan" address table pointer.
DVDTblB	dw	(DVD1Tbl-@)	;CD/DVD setup address table pointer.
SFlag	db	4		;User cache-size flag (default = 4).
BiosHD	db	0FFh		;Number of BIOS hard disks.
HDUnit	db	0		;Current BIOS unit number.
	db	0		;(Unused alignment "filler").
;
; CD/DVD Unit Parameter Tables.
;
UTable	dw	0FFFFh		;Unit 0 DMA address (set by Init).
	dw	0FFFFh		;	IDE address (set by Init).
	dw	0FFFFh		;	(Unused alignment "filler").
	db	0FFh		;	Device-select (set by Init).
	db	0FFh		;	Media-change flag.
	dd	0FFFFFFFFh	;	Current audio-start address.
	dd	0FFFFFFFFh	;	Current audio-end   address.
	dd	0FFFFFFFFh	;	Last-session starting LBA.
	dw	0FFFFh		;Unit 1 DMA address (set by Init).
	dw	0FFFFh		;	IDE address (set by Init).
	dw	0FFFFh		;	(Unused alignment "filler").
	db	0FFh		;	Device-select (set by Init).
	db	0FFh		;	Media-change flag.
	dd	0FFFFFFFFh	;	Current audio-start address.
	dd	0FFFFFFFFh	;	Current audio-end   address.
	dd	0FFFFFFFFh	;	Last-session starting LBA.
	dw	0FFFFh		;Unit 2 DMA address (set by Init).
	dw	0FFFFh		;	IDE address (set by Init).
	dw	0FFFFh		;	(Unused alignment "filler").
	db	0FFh		;	Device-select (set by Init).
	db	0FFh		;	Media-change flag.
	dd	0FFFFFFFFh	;	Current audio-start address.
	dd	0FFFFFFFFh	;	Current audio-end   address.
	dd	0FFFFFFFFh	;	Last-session starting LBA.
	dw	0FFFFh		;Unit 3 DMA address (set by Init).
	dw	0FFFFh		;	IDE address (set by Init).
	dw	0FFFFh		;	(Unused alignment "filler").
	db	0FFh		;	Device-select (set by Init).
	db	0FFh		;	Media-change flag.
	dd	0FFFFFFFFh	;	Current audio-start address.
	dd	0FFFFFFFFh	;	Current audio-end   address.
	dd	0FFFFFFFFh	;	Last-session starting LBA.
UTblEnd	equ	$		;(End of all unit tables).
;
; Global Descriptor Table Pointer, used by our "lgdt" command.
;
GDTP	dw	GDTLen		;GDT length (always 24 bytes).
GDTPAdr	dd	(OurGDT-@)	;GDT 32-bit address (Set By Init).
RqIndex	dw	(Ctl1Tbl-@)	;Request cache-table index, here
				;  for "OurGDT" 32-bit alignment.
;
; Global Descriptor Table, used during our own "real-mode" moves.
;
OurGDT	dd	0,0		;"Null" segment to start.
	dw	0FFFFh		;Conforming code segment.
GDT_CSL	dw	0
GDT_CSM	db	0
	dw	0009Fh
GDT_CSH	db	0
GDT_DS	dd	00000FFFFh,000CF9300h  ;4-GB data segment in pages.
GDTLen	equ	($-OurGDT)
;
; Int 15h Descriptor Table, used for BIOS "protected-mode" moves.
;
MoveDT	dd	0,0		       ;"Null" descriptor.
	dd	0,0		       ;GDT descriptor.
SrcDsc	dd	00000FFFFh,000009300h  ;Source segment descriptor.
DstDsc	dd	00000FFFFh,000009300h  ;Dest.  segment descriptor.
	dd	0,0		       ;(Used by BIOS).
	dd	0,0		       ;(Used by BIOS).
;
; DOS Device-Interrupt Routine.   This logic handles CD/DVD requests.
;   Int 13h I-O requests for BIOS units are handled at "Entry" below.
;
DevInt:	pushf			;Entry -- save CPU flags.
	cli			;Disable CPU interrupts.
	pushad			;Save all CPU registers.
	push	ds
	push	es
	bts	cs:IOF.lb,7	;Is this driver currently "busy"?
	jc s	DvIBsy		;Yes?  Post "general error" and exit!
	mov	ah,005h		;Issue "A20 local-enable" request.
	call	A20Req		;(Makes critical settings on exit).
	jnz s	DvIA2E		;If "A20" failure, use routine below.
	mov	ah,20		;Get CD/DVD table-length multiplier.
	les	si,[bx+RqPkt-@]	;Point to DOS request packet.
	db	0EAh		;Go to main CD/DVD request handler.
@CDMain	dd	(CDReq-@)	;(CD/DVD request address, Init set).
DvIRtn:	call	VDUnlk		;If needed, VDS "unlock" user buffer.
	mov	ah,006h		;Issue "A20 local-disable" request.
	call	A20Req
DvIntE:	cli			      ;Disable CPU interrupts.
	and	[bx+IOF-@].lw,00001h  ;Zero all but cache-flush flag.
DvIntX:	pop	es		      ;Reload all CPU registers.
	pop	ds
	popad
	popf			;Reload CPU flags and exit.
	retf
DvIBsy:	call	DvIErr		;We're BUSY, you FOOL!  Post "general
	jmp s	DvIntX		;  error" in request packet and exit!
DvIA2E:	call	DvIErr		;"A20 enable" FAILURE!  Post "general
	jmp s	DvIntE		;  error" in request packet and exit!
DvIErr:	lds	si,cs:RqPkt	    ;Error handler -- Post "general
	mov	[si].RPStat,0810Ch  ;  error" in caller's packet.
	ret			    ;Exit.
;
; Main Hard-Disk and Diskette Int 13h Entry Routine.   For a CHS
;   request, the registers are:
;
;   AH      Request code.  We handle 002h read and 003h write.
;   AL      I-O sector count.
;   CH      Lower 8 bits of starting cylinder.
;   CL      Starting sector and upper 2 bits of cylinder.
;   DH      Starting head.
;   DL      BIOS unit number:  000h/001h diskettes, 080h+ hard-disks.
;   ES:BX   I-O buffer address.
;
; For an LBA request, the registers are:
;
;   AH      Request code.  We handle 042h read and 043h write.
;   DL      BIOS unit number:  000h/001h diskettes, 080h+ hard-disks.
;   DS:SI   Pointer to Device Address Packet ("DAP") as shown above.
;
Entry:	pushf			;Save CPU flags and BP-reg.
	push	bp
	call	Switch		;Switch to this driver's stack.
	mov	bp,0		;Reset active-units table index.
@LastU	equ	[$-2].lw	;(Last-unit index, set by Init).
NextU:	dec	bp		;Any more active units to check?
	js s	QuickX		;No, not our request -- exit quick!
	cmp	dl,[bp+Units-@]	;Does request unit match our table?
	jne s	NextU		;No, see if more units remain.
	mov	dl,0BEh		;Mask out LBA & write request bits.
	and	dl,ah
	cmp	dl,002h		;Is this a CHS or LBA read or write?
	mov	dl,[bp+TypeF-@]	;(Set unit "type" flag in any case).
	mov	cs:RqTyp,dl
	jne s	NotUs		;No, not for us -- Check diskette req.
	shl	ah,1		;Is this an LBA read or write request?
	jns s	ValSC		;No, go validate CHS sector count.
	cmp	[si].DapBuf,-1	;Is this a 64-bit "DAP" I-O buffer?
	jne s	GetDAP		;No, go get all "DAP" parameters.
NotUs:	cmp	dl,0C0h		;Not for us -- Diskette I-O request?
	jne s	Pass		;No, "flush" cache and pass request.
	cmp	ah,005h		;Diskette track-format request?
	jne s	QuickX		;No, "pass" other diskette requests.
	nop			;(Unused alignment "filler").
Pass:	or	cs:IOF.lb,001h	;Flush cache on next caching request.
QuickX:	cli			;Pass to BIOS -- disable interrupts.
	and	cs:IOF.lb,001h	;Reset "busy" but save "flush" flag.
	pop	es		;Reload all CPU registers.
	pop	ds
	popad
	lss	sp,cs:CStack	;Switch back to caller's stack.
	pop	bp		;Reload BP-register and CPU flags.
	popf
	db	0EAh		;Go to next routine in Int 13h chain.
@Prv13A	dd	0		;(Prior Int 13h vector, set by Init).
GetDAP:	mov	al,[si].DapSC	;Get "DAP" I-O sector count.
	les	dx,[si].DapLBA1	;Get "DAP" LBA bits 16-47.
	mov	di,es
	les	cx,[si].DapBuf	;Get "DAP" I-O buffer address.
	mov	si,[si].DapLBA	;(Get "DAP" LBA bits 0-15).
ValSC:	dec	al		;Is sector count zero or over 128?
	js s	Pass		;Yes?  Let BIOS handle this "No-No!".
	inc	ax		;Restore sector count -- LBA request?
	js s	SetDS		;Yes, go set driver's DS-reg.
	xchg	ax,cx		;CHS -- save request code and sectors.
	xor	di,di		;Reset upper LBA address bits.
	mov	si,0003Fh	;Set SI-reg. to starting sector.
	and	si,ax
	dec	si
	shr	al,6		;Set AX-reg. to starting cylinder.
	xchg	al,ah
	xchg	ax,dx		;Swap cylinder & head values.
	mov	al,[bp+CHSec-@] ;Get disk CHS sectors/head value.
	or	al,al		;Were disk CHS values legitimate?
	jz s	Pass		;No?  Let BIOS handle this request!
	push	ax		;Save CHS sectors/head value.
	mul	ah		;Convert head to sectors.
	add	si,ax		;Add result to starting sector.
	pop	ax		;Reload CHS sectors/head value.
	mul	[bp+CHSHd-@].lb ;Convert cylinder to sectors.
	mul	dx
	add	si,ax		;Add to head/sector value.
	adc	dx,di
	xchg	ax,bx		;Get buffer offset in AX-register.
	xchg	ax,cx		;Swap offset with request/sectors.
SetDS:	push	cs		;Set this driver's DS-reg.
	pop	ds
	xor	bx,bx		;Zero BX-reg. for relative logic.
	and	ah,006h		;Save sector count & command bits.
	mov	[bx+RqSec-@],ax
	mov	[bx+VDSOf-@].lw,cx   ;Set VDS I-O buffer address.
	mov	[bx+VDSOf-@].hw,bx
	mov	[bx+VDSSg-@],es
	mov	ah,0		     ;Set VDS buffer length.
	shl	ax,1
	mov	[bx+VDSLn+1-@],ax
	push	cs		     ;Do VDS "lock" on user buffer.
	call	VDLock
	jc s	Pass		     ;VDS error:  Pass this request!
	cmp	[bx+RqTyp-@].lb,0C0h ;Is this a diskette request?
	jne s	SetBuf		     ;No, go set user buffer address.
	mov	es,bx		     ;Point ES-reg. to low memory.
	mov	cl,01Fh		     ;Get BIOS diskette status code.
	and	cl,es:[DKTSTAT]
	cmp	cl,MCHANGE	   ;Diskette media-change detected?
	jne s	SetBuf		   ;No, go set 32-bit buffer address.
	or	[bx+IOF-@].lb,1	   ;Flush cache before next request.
SetBuf:	mov	[bx+RqBuf-@],eax   ;Set user I-O buffer address.
	mov	[bx+RqLBA-@],di	   ;Set initial request LBA.
	mov	[bx+RqLBA+2-@],dx
	mov	[bx+RqLBA+4-@],si
	mov	[bx+RqUNo-@],bp	   ;Set "cache unit" number.
	mov	ah,005h		   ;Issue "A20 local-enable" request.
	call	A20Req
	jnz s	A20Er		;If "A20" error, use routine below.
GoMain:	db	0EAh		;Go to main cache or driver routine.
@HDMain	dd	(Cache-@)	;(Main entry address, set by Init).
HDDRtn:	pushf			;Save error flag (carry) & error code.
	push	ax
	mov	ah,006h		;Issue "A20 local-disable" request.
	call	A20Req
Abandn:	call	VDUnlk		;Do VDS "unlock" of user I-O buffer.
	pop	ax		;Reload error code and error flag.
	popf
	mov	bp,sp		;Point to saved registers on stack.
	mov	[bp+33],al	;Set error code in exiting AH-reg.
	cli			;Disable CPU interrupts.
	rcl	[bx+IOF-@].lb,1	;Reset "busy" but save "flush" flag.
	shr	[bx+IOF-@].lb,1
	pop	es		;Reload all CPU registers.
	pop	ds
	popad
	lss	sp,cs:CStack	;Switch back to caller's stack.
GetOut:	mov	bp,sp		;Set error flag in exiting carry bit.
	rcr	[bp+8].lb,1	;(One of the flags saved by Int 13h).
	rol	[bp+8].lb,1
	pop	bp		;Reload BP-register and CPU flags.
	popf
	iret			;Exit.
A20Er:	or	[bx+IOF-@].lb,1	;"A20" error!  Request cache "flush".
	stc			;Stack error flag (carry bit).
	pushf
	push	ax		;Stack "XMS error" code (AL = 0FFh).
	jmp s	Abandn		;Abandon this request and exit QUICK!
InvalF:	mov	ah,001h		;We are BUSY, you fool!  Set "invalid
	jmp s	GetOut		;  function" code and get out, QUICK!
;
; Subroutine to check for "busy" and switch to this driver's stack.
;
Switch:	cli			;Disable CPU interrupts.
	bts	cs:IOF,7	;Is this driver currently "busy"?
	pop	bp		;(Load return address into BP-reg.).
	jc	InvalF		;Yes?  Set "invalid function" & exit!
	mov	cs:CStack.lw,sp	;Save caller's stack pointer.
	mov	cs:CStack.hw,ss
	push	cs		;Switch to this driver's stack.
	pop	ss
	mov	sp,0
@Stack1	equ	[$-2].lw	;(Starting stack ptr., set by Init).
	sti			;Re-enable CPU interrupts.
	cld			;Ensure FORWARD "string" commands.
	pushad			;Save all CPU registers.
	push	ds
	push	es
	jmp	bp		;Exit.
CStack	dd	0		;Caller's saved stack pointer, here
				;  for 8-byte alignment of "VDLock".
;
; Subroutine to issue a VDS "lock" request.
;
VDLock:	pusha			     ;Save all 16-bit CPU registers.
	push	ds
	or	[bx+VDSAd-@].dwd,-1  ;Invalidate VDS buffer address.
	mov	ax,08103h	     ;Get VDS "lock" parameters.
	mov	dx,0000Ch
	call	VDReq		     ;Do VDS "lock" on user buffer.
	pop	ds		     ;Reload all 16-bit registers.
	popa
	jc s	VDLokX		     ;VDS error:  Exit immediately!
	mov	eax,[bx+VDSAd-@]     ;Get 32-bit VDS buffer address.
	cmp	eax,-1		     ;Is VDS address not "all-ones"?
	jb s	VDLokF		     ;Yes, set VDS "lock" flag.
	movzx	eax,[bx+VDSSg-@].lw  ;VDS logic is NOT present --
	shl	eax,4		     ;  set 20-bit buffer address.
	add	eax,[bx+VDSOf-@]
	mov	[bx+VDSAd-@],eax
	nop			     ;(Unused alignment "filler").
VDLokF:	adc	[bx+VLF-@],bl	     ;Set VDS "lock" flag from carry.
VDLokX:	retf			     ;Exit.
;
; DOS "Strategy" Routine, used only for CD/DVD requests.
;
Strat:	mov	cs:RqPkt.lw,bx	;Save DOS request-packet address.
	mov	cs:RqPkt.hw,es
	retf			;Exit & await DOS "Device Interrupt".
	db	0		;(Unused alignment "filler").
;
; "External Drive" Entry Routine.    This logic sets up a read/write
;   request for handling at "SetBuf"above.   At entry, the registers
;   contain:
;
;   EAX     32-bit user buffer address.    MUST be valid, and must
;	      already have been "VDS locked" if needed.
;   DS:BX   "Call back" I-O routine address, used to do I-O for an
;	      LBA address that is NOT found to be "cached".   NOTE
;	      that the upper half of the EBX-reg. will be LOST!
;   CL      I-O sector count (1 to 128) of 512-byte sectors.
;   CH      000h for a read request, 002h for a write request.
;   SI      LBA address bits 0-15.
;   DX      LBA address bits 16-31.
;   DI      LBA address bits 32-47.
;   BP      Bits 0-7 are the "cache unit" number, MAXBIOS through
;	      MAXBIOS+3 for CD/DVD drives 0 to 3 handled by UDVD,
;	      and MAXBIOS+4 or higher for other external drivers.
;	      Bits 8-15 are ignored.
;
ExEntr:	push	bx		;Save unit no. in lower BX-reg. and
	push	bp		;  "call back" addr. in upper BX-reg.
	pop	ebx
	pushf			;Save CPU flags and BP-reg.
	push	bp
	call	Switch		;Switch to this driver's stack.
	push	ds		;Stack "call back" address & unit no.
	push	ebx
	call	VDRSet			;Do critical driver settings.
	pop	bp			;Set desired unit in BP-reg.
	pop	ExtDvr			;Set "call back" routine ptr.
	mov	[bx+RqSec-@],cx		;Set sectors and I-O command.
	mov	[bx+RqTyp-@].lb,EXTYPE	;Post "external" device type.
	jmp	SetBuf			;Go post other data & do I-O.
	db	0			;(Unused alignment "filler").
;
; "External Reset" Routine, MUST be 32 bytes after "ExEntr" above!
;
ExRSet:	or	cs:IOF.lb,001h	;Flush cache on next caching request.
	retf			;Exit.
	db	0		;(Unused alignment "filler").
;
; Subroutine to move our DMA command-list up to XMS memory for I-O.
;   This avoids "DMA Hangs" if we are loaded in UMBPCI "Shadow RAM"!
;
MovCLX:	mov	ecx,8		 ;Set command-list move parameters
	mov	esi,(IOAdr-@)	 ;  and then fall thru to "MvData".
@CLAddr	equ	[$-4].dwd	 ;(Command-list address, Init set).
	mov	edi,[bx+PRDAd-@]
;
; Subroutine to do XMS moves, MUST be 56 bytes after "ExEntr" above!
;   At entry, the ECX-reg. has the byte length and the ESI/EDI-regs.
;   have the move addresses.    At exit, the carry flag is reset for
;   no error or SET for any XMS move error.   If so, the AL-reg. has
;   a 0FFh XMS error code.   The DS- and ES-regs. are saved/restored
;   and the BX-reg. is zeroed at exit.   The other registers are NOT
;   saved, for speed.   This subroutine must be called with a 32-bit
;   "call far" command.    Note that a "real mode" move posts no XMS
;   errors!   Only "protected mode" moves (BIOS/EMM386) post errors.
;
MvData:	push	ds		;Save caller's segment registers.
	push	es
	push	cs		;Set this driver's data segment.
	pop	ds
	smsw	ax		;Get CPU "machine status word".
	shr	ax,1		;Are we running in "protected" mode?
	jc s	MvBIOS		;Yes, use BIOS move routine below.
	shr	ecx,2		;Convert byte count to dword count.
	mov	bp,cx		;Set move byte count in BP-reg.
	mov	dx,512		;Set 512-dword section count.
MvRNxt:	mov	cx,dx		;Get move section count.
	cmp	cx,bp		;At least 512 dwords left?
	jbe s	MvRGo		;Yes, use full section count.
	mov	cx,bp		;Use remaining dword count.
MvRGo:	cli			;Disable CPU interrupts.
	db	02Eh,00Fh	;"lgdt cs:GDTP", coded the hard-way
	db	001h,016h	;   to avoid any annoying V5.1 MASM
	dw	(GDTP-@)	;   warning messages about it!
	mov	eax,cr0		;Set "protected-mode" control bit.
	or	al,001h
	mov	cr0,eax
	mov	bx,(GDT_DS-OurGDT)  ;Set DS- and ES-reg. "selectors".
	mov	ds,bx
	mov	es,bx
	cld			;Ensure FORWARD "string" commands.
	db	0F3h,067h	;Move all 32-bit words using "rep"
	movsd			;  and "address-override" prefixes.
	db	067h,090h	;("Override" & "nop" for 386 BUG!).
	dec	ax		;Clear "protected-mode" control bit.
	mov	cr0,eax
	sti			;Allow interrupts after next command.
	sub	bp,dx		;Any more data sections to move?
	ja s	MvRNxt		;Yes, go move next data section.
MvDone:	xor	ax,ax		;Success!  Reset carry and error code.
MvExit:	sti			;RESTORE all critical driver settings!
	cld			;(Never-NEVER "trust" external code!).
	pop	es		;Reload caller's segment registers.
	pop	ds
	mov	bx,0		;Rezero BX-reg. but save carry bit.
	retf			;Exit ("retf" for calls from UDVD).
MvBIOS:	mov	edx,ecx		;BIOS move:  Set length in EDX-reg.
	xor	ecx,ecx		;Initialize 16K section count the
	mov	ch,040h		;  "5-byte" way (helps alignment!).
MvBNxt:	push	ecx		;Save all 32-bit move parameters.
	push	edx
	push	esi
	push	edi
	cmp	ecx,edx		;At least 1 full section left?
	jbe s	MvBGo		;Yes, use full section count.
	mov	ecx,edx		;Use remaining data byte count.
MvBGo:	cld			;Ensure FORWARD "string" commands!
	shr	cx,1		;Convert byte count to word count.
	push	ds		;Ensure ES-reg. points to our driver.
	pop	es
	push	edi		;Set up destination descriptor.
	mov	di,(DstDsc+2-@)
	pop	ax
	stosw
	pop	ax
	stosb
	mov	[di+2],ah
	push	esi		      ;Set up source descriptor ("sub"
	sub	di,(DstDsc-SrcDsc+3)  ;  will zero carry for Int 15h).
	pop	ax
	stosw
	pop	ax
	stosb
	mov	[di+2],ah
	mov	si,(MoveDT-@)	;Point to move descriptor table.
	mov	ah,087h		;Have BIOS move next data section.
	int	015h
	pop	edi		;Reload all 32-bit move parameters.
	pop	esi
	pop	edx
	pop	ecx
	mov	al,XMSERR	;Get our XMS error code.
	jc s	MvExit		;If any BIOS error, exit immediately.
	sti			;Ensure CPU interrupts are enabled.
	push	cs		;Ensure DS-reg. points to our driver.
	pop	ds
	add	esi,ecx		;Update source and dest. addresses.
	add	edi,ecx
	sub	edx,ecx		;Any more data sections to move?
	ja s	MvBNxt		;Yes, go move next section of data.
	jmp s	MvDone		;Success!  Reset carry flag and exit.
;
; Subroutine to issue "A20" local-enable and local-disable requests.
;
A20Req:	db	09Ah		;Call XMS manager for "A20" request.
@XEntry	dd	0		;(XMS "entry" address, set by Init).
	dec	ax		;Zero AX-reg. if success, -1 if error.
	jmp s	VDRSet		;Go restore critical settings & exit.
;
; Subroutine to issue a VDS "unlock" request.
;
VDExit:	ret			;(Aligned "ret", for faster speed).
VDUnlk:	shr	[bx+VLF-@].lb,1	;"Unlock" -- Was I-O buffer "locked"?
	jnc s	VDExit		;No, just exit above.
	mov	ax,08104h	;Get VDS "unlock" parameters.
	mov	dx,0
VDReq:	mov	di,(VDSLn-@)	;Point to VDS parameter block.
	push	ds
	pop	es
	int	04Bh		;Execute desired VDS request.
	nop			;(Unused alignment "filler").
VDRSet:	sti			;RESTORE all critical driver settings!
	cld			;(Never-NEVER "trust" external code!).
	push	cs
	pop	ds
	mov	bx,0		;Rezero BX-reg. but save carry bit.
	ret			;Exit.
NormSz1	equ	$+STACK		;Size of "small" normal-memory logic.
				;  All logic below may go in the HMA.
;
; Main CD/DVD request handler.
;
CDReq:	call	ClrPkt		;Clear our ATAPI packet area.
	mov	[bx+XMF-@],bx	       ;Reset XMS & DMA-input flags.
	mov	es:[si.RPStat],RPDON   ;Set packet status to "done".
	mov	al,es:[si+RPSubU].lb   ;Get our CD/DVD drive number.
	mul	ah		       ;Get drive's "audio-start" ptr.
	add	ax,(UTable+8-@)
	xchg	ax,di
	mov	[bx+AudAP-@],di
	mov	eax,[di-8]	       ;Set drive DMA/IDE addresses.
	mov	[bx+DMAAd-@],eax
	mov	al,es:[si+RPOp]	       ;Get packet request code.
	mov	di,offset (DspTbl1-@)  ;Point to 1st dispatch table.
@DspTB1	equ	[$-2].lw	       ;(Dispatch address, Init set).
	call	Dispat		;Dispatch to desired request handler.
	db	0EAh		;Return to CD/DVD exit logic.
	dw	(DvIRtn-@)
@CDXSeg	dw	0		;(Driver segment address, Init set).
;
; Subroutine to initialize the next UltraDMA I-O transfer.
;   This routine is located here for 8-byte alignment.
;
IniDMA:	mov	dx,[bx+DMAAd-@]	;Get DMA command-register address.
	mov	si,(PRDAd-@)	;Get Physical-Region Descriptor addr.
	out	dx,al		;Reset DMA commands and set DMA mode.
	inc	dx		;Point to DMA status register.
	inc	dx
	in	al,dx		;Reset DMA status register.
	or	al,006h		;(Done this way so we do NOT alter
	out	dx,al		;  the "DMA capable" status flags!).
	inc	dx		;Point to PRD address register.
	inc	dx
	outsd			;Set PRD pointer to our DMA address.
	retf			;Exit.
;
; CD/DVD Dispatch Table for DOS request codes 0 through 14.   Each
;   entry in the following tables has 4 bits of "IOCTL byte count"
;   and 12 bits of offset for its "handler" routine, so "handlers"
;   MUST be located within 4K bytes from the start of this driver!
;
DspTbl1	dw	DspLmt1		;Number of valid request codes.
	dw	Try2D-USR	;Invalid-request handler address.
DspTblA	dw	0		;00 -- Init  (handled by I_Init).
	dw	0		;01 -- Media Check	(unused).
	dw	0		;02 -- Build BPB	(unused).
	dw	Try3D-USR	;03 -- IOCTL Input.
	dw	0		;04 -- Input		(unused).
	dw	0		;05 -- Input no-wait	(unused).
	dw	0		;06 -- Input Status	(unused).
	dw	0		;07 -- Input flush	(unused).
	dw	0		;08 -- Output		(unused).
	dw	0		;09 -- Output & verify	(unused).
	dw	0		;10 -- Output status	(unused).
	dw	0		;11 -- Output flush	(unused).
	dw	Try4D-USR	;12 -- IOCTL Output.
	dw	RqExit-USR	;13 -- Device Open     (ignored).
	dw	RqExit-USR	;14 -- Device Close    (ignored).
DspLmt1	equ	($-DspTblA)/2	;Table request-code limit.
;
; CD/DVD Dispatch Table for DOS request codes 128 through 136.
;
DspTbl2	dw	DspLmt2		;Number of valid request codes.
	dw	0		;Invalid-request handler address.
DspTblB	dw	RqRead-USR	;128 -- Read Long.
	dw	0		;129 -- Reserved	(unused).
	dw	RqSeek-USR	;130 -- Read Long Prefetch.
	dw	RqSeek-USR	;131 -- Seek.
	dw	RqPlay-USR	;132 -- Play Audio.
	dw	RqStop-USR	;133 -- Stop Audio.
	dw	0		;134 -- Write Long	(unused).
	dw	0		;135 -- Wr. Long Verify	(unused).
	dw	RqRsum-USR	;136 -- Resume Audio.
DspLmt2	equ	($-DspTblB)/2	;Table request-code limit.
;
; CD/DVD Dispatch table for IOCTL Input requests.
;
DspTbl3	dw	DspLmt3		    ;Number of valid request codes.
	dw	0		    ;Invalid-request handler address.
DspTblC	dw	ReqDHA+(5*IXM)-USR  ;00 -- Device-header address.
	dw	RqHLoc+(6*IXM)-USR  ;01 -- Current head location.
	dw	0		    ;02 -- Reserved	    (unused).
	dw	0		    ;03 -- Error Statistics (unused).
	dw	0		    ;04 -- Audio chan. info (unused).
	dw	0		    ;05 -- Read drive bytes (unused).
	dw	ReqDS +(5*IXM)-USR  ;06 -- Device status.
	dw	ReqSS +(4*IXM)-USR  ;07 -- Sector size.
	dw	ReqVS +(5*IXM)-USR  ;08 -- Volume size.
	dw	ReqMCS+(2*IXM)-USR  ;09 -- Media-change status.
	dw	ReqADI+(7*IXM)-USR  ;10 -- Audio disk info.
	dw	ReqATI+(7*IXM)-USR  ;11 -- Audio track info.
	dw	ReqAQI+(11*IXM)-USR ;12 -- Audio Q-channel info.
	dw	0		    ;13 -- Subchannel info  (unused).
	dw	0		    ;14 -- Read UPC code    (unused).
	dw	ReqASI+(11*IXM)-USR ;15 -- Audio status info.
DspLmt3	equ	($-DspTblC)/2	    ;Table request-code limit.
;
; CD/DVD Dispatch table for IOCTL Output requests.
;
DspTbl4	dw	DspLmt4		    ;Number of valid request codes.
	dw	0		    ;Invalid-request handler address.
DspTblD	dw	ReqEJ +(1*IXM)-USR  ;00 -- Eject Disk.
	dw	ReqLU +(2*IXM)-USR  ;01 -- Lock/Unlock Door.
	dw	ReqRS +(1*IXM)-USR  ;02 -- Reset drive.
	dw	0		    ;03 -- Audio control    (unused).
	dw	0		    ;04 -- Write ctl. bytes (unused).
	dw	ReqCl +(1*IXM)-USR  ;05 -- Close tray.
DspLmt4	equ	($-DspTblD)/2	    ;Table request-code limit.
;
; CD/DVD Function-Code "Dispatch" Routines.
;
USR:	mov	al,3		;Unsupported request!  Get error code.
ReqErr:	les	si,[bx+RqPkt-@]	;Reload DOS request-packet address.
	mov	ah,081h		;Post error flags & code in packet.
	mov	es:[si+RPStat],ax
RqExit:	ret			;Exit ("ignored" request handler).
Try2D:	sub	al,080h		;Not request code 0-15:  subtract 128.
	mov	di,(DspTbl2-@)	;Point to 2nd DOS dispatch table.
@DspTB2	equ	[$-2].lw	;(Dispatch table address, Init set).
	jmp s	Dispat		;Go try request-dispatch again.
	db	0		;(Unused alignment "filler").
Try3D:	mov	di,(DspTbl3-@)	;Point to IOCTL Input dispatch table.
@DspTB3	equ	[$-2].lw	;(Dispatch table address, Init set).
	jmp s	TryIOC		;Go get actual IOCTL request code.
Try4D:	mov	di,(DspTbl4-@)	;Point to IOCTL Output table.
@DspTB4	equ	[$-2].lw	;(Dispatch table address, Init set).
TryIOC:	les	si,es:[si+IOCAdr]  ;Get actual IOCTL request code.
	mov	al,es:[si]
	les	si,[bx+RqPkt-@]	;Reload DOS request-packet address.
Dispat:	cmp	al,cs:[di]	;Is request code out-of-bounds?
	inc	di		;(Skip past table-limit value).
	inc	di
	jae s	Dspat1		;Yes?  Dispatch to error handler!
	inc	di		;Skip past error-handler address.
	inc	di
	xor	ah,ah		;Point to request-handler address.
	shl	ax,1
	add	di,ax
Dspat1:	mov	dx,cs:[di]	;Get handler address from table.
	mov	di,00FFFh
	and	di,dx
	xor	dx,di		;IOCTL request (xfr length > 0)?
	jz s	DspGo		;No, dispatch to desired handler.
	shr	dx,12		   ;Ensure correct IOCTL transfer
	mov	es:[si+IOCLen],dx  ;  length is set in DOS packet.
	les	si,es:[si+IOCAdr]  ;Get IOCTL data-transfer address.
DspGo:	add	di,(USR-@)	   ;Dispatch to desired handler.
@DspOfs	equ	[$-2].lw	   ;(HMA dispatch offset, Init set).
	jmp	di
;
; CD/DVD subroutine to validate starting RedBook disk sector numbers.
;
ValSN:	mov	eax,es:[si+RLSec]  ;Get starting sector number.
ValSN1:	mov	dl,es:[si+RLAM]	;Get desired addressing mode.
	cmp	dl,001h		;HSG or RedBook addressing?
	ja s	ValSNE		;No?  Return "sector not found".
	jne s	RqExit		;HSG -- exit (accept any DVD value).
	call	CvtLBA		;RedBook -- get starting sector.
	cmp	eax,RMAXLBA	;Is starting sector too big?
	jbe s	RqExit		;No, all is well -- go exit above.
ValSNE:	pop	ax		;Error!  Discard our exit address.
SectNF:	mov	al,8		;Sector not found!  Get error code.
	jmp s	ReqErr		;Go post "sector not found" and exit.
;
; CD/DVD "Read Long" request handler.   All CD/DVD data is read here.
;
RqRead:	call	ValSN		;Validate starting sector number.
	call	MultiS		;Handle Multi-Session disk if needed.
	jc s	ReqErr		;If error, post return code & exit.
	mov	cx,es:[si+RLSC]	;Get request sector count.
	jcxz	RqExit		;If zero, simply exit.
	xchg	cl,ch		;Save swapped sector count.
	mov	[bx+PktLn-@],cx
	cmp	es:[si.RLDM],1	;"Cooked" or "raw" read mode?
	ja s	SectNF		;Neither?  Return "sector not found"!
	mov	dl,028h		;Get "cooked" input values.
	mov	ax,COOKSL
	jb s	RqRL1		;If "cooked" input, set values.
	mov	dl,0BEh		;Get "raw" input values.
	mov	ax,RAWSL
	mov	[bx+PktRM-@].lb,0F8h  ;Set "raw" input flags.
RqRL1:	mov	[bx+Packet-@].lb,dl   ;Set "packet" opcode.
	mul	es:[si+RLSC]	      ;Get desired input byte count.
	test	dx,dx		      ;More than 64K bytes desired?
	jnz s	SectNF		      ;Yes?  Return sector not found!
	mov	[bx+VDSLn-@],ax	      ;Set VDS and DMA byte counts.
	mov	[bx+IOLen-@],ax
	les	ax,es:[si+RLAddr]     ;Set user input-buffer address.
	mov	[bx+VDSOf-@],ax
	mov	[bx+VDSSg-@],es
	mov	[bx+UserB+2-@],es
	db	09Ah		      ;Do VDS "lock" on user buffer.
	dw	(VDLock-@)
@RQLSeg	dw	0		      ;Driver segment, set by Init).
	les	si,[bx+RqPkt-@]	      ;Reload request packet address.
	jc s	RqRL5		      ;"Lock" error:  Use "PIO mode".
	test	[bx+DMAAd-@].lb,001h  ;Is this drive using UltraDMA?
	jnz s	RqRL2		      ;No, check for "raw mode" read.
	nop			      ;(Unused alignment "filler").
	push	cs		      ;Set 32-bit DMA buffer address.
	call	SetAdr
	adc	[bx+XMF-@].lw,00100h  ;Set XMS-move & DMA-input flags.
RqRL2:	cmp	[bx+PktRM-@],bl	      ;Is this a "raw mode" read?
@NoCA1:	jne s	RqRL6		      ;Yes, use "non-cached" logic.
				      ;("jmp s RqRL6" if no caching).
	movzx	ax,es:[si+RLSC].lb    ;Get sector count & read code.
	shl	al,2		      ;Multiply by 4 for 2K sectors.
	mov	[bx+RqSec-@],ax	      ;Set sectors and I-O command.
	movzx	bp,es:[si+RPSubU].lb  ;Get our CD/DVD drive number.
	add	bp,MAXBIOS	      ;Convert to our "cache unit".
	mov	eax,[bx+PktLBA-@]     ;Convert packet LBA to "little
	call	Swp32		      ;  endian" format for caching.
	shl	eax,2		      ;Multiply by 4 for 2K sectors.
	push	eax		      ;Put 32-bit LBA in DX/SI-regs.
	pop	si		      ;  and zero DI-reg. (hi-order)
	pop	dx		      ;  as CD/DVD drives do not use
	xor	di,di		      ;  48-bit LBAs!
	mov	eax,[bx+VDSAd-@]        ;Get 32-bit input buffer addr.
	mov	[bx+RqTyp-@].lb,CDTYPE	;Set our CD/DVD device type.
	pushf				;Enter Int 13h caching logic.
	push	cs
	call	CachCD
	jnc s	RqRLX		;Back from caching!  Any read errors?
	mov	al,ah		;Yes, get error code in AL-reg.
RqRL3:	cmp	al,0FFh		;Did caching post "XMS error"?
	jne s	RqRL4		;No, set returned error code as-is.
	mov	al,GENERR	;Get CD/DVD "general error" code.
RqRL4:	jmp	ReqErr		;Set error code in packet & exit.
RqRL5:	or	[bx+IOF-@].lb,1	;Flush cache on next request.
RqRL6:	push	cs		;Noncached -- Input desired data.
	call	RLRead
	jc s	RqRL3		;If any error, handle it above.
RqRLX:	ret			;All is well -- Exit.
;
; Subroutine to enter our caching logic for CD/DVD input requests.
;
CachCD:	pushf			  ;Save CPU flags & BP-reg. for exit.
	push	bp
	cli			;Disable CPU interrupts.
	mov	CStack.lw,sp	;Save current stack pointer.
	mov	CStack.hw,ss
	push	ds		;Switch to this driver's stack.
	pop	ss
	mov	sp,0
@Stack2	equ	[$-2].lw	;(Starting stack ptr., set by Init).
	sti			;Re-enable CPU interrupts.
	cld			;Ensure FORWARD "string" cmds.
	pushad			;Save all CPU registers.
	push	ds
	push	es
	db	0EAh		;Use caching logic for CD/DVD input.
	dw	(SetBuf-@)
@CCDSeg	dw	0		;(Driver segment address, Init set).
	db	0		;(Unused alignment "filler").
;
; CD/DVD "Seek" request handler.
;
RqSeek:	call	RdAST1		;Read current "audio" status.
	call	ClrPkt		;Reset our ATAPI packet area.
	jc s	RqSK2		;If status error, do DOS seek.
	mov	al,[di+1]	;Get "audio" status flag.
	cmp	al,011h		;Is drive in "play audio" mode?
	je s	RqSK1		;Yes, validate seek address.
	cmp	al,012h		;Is drive in "pause" mode?
	jne s	RqSK2		;No, do DOS seek below.
RqSK1:	call	ValSN		;Validate desired seek address.
	mov	di,[bx+AudAP-@]	;Point to audio-start address.
	cmp	eax,[di+4]	;Is address past "play" area?
	ja s	RqSK2		;Yes, do DOS seek below.
	mov	[di],eax	;Update audio-start address.
	call	PlayA		;Issue "Play Audio" command.
	jc s	RqPLE		;If error, post return code & exit.
	cmp	[di+1].lb,011h	;Were we playing audio before?
	je s	RqPLX		;Yes, post "busy" status and exit.
	call	ClrPkt		;Reset our ATAPI packet area.
	jmp s	RqStop		;Go put drive back in "pause" mode.
RqSK2:	call	ValSN		;Validate desired seek address.
	call	MultiS		;Handle Multi-Session disk if needed.
	jc s	RqPLE		;If error, post return code & exit.
	mov	[bx+Packet-@].lb,02Bh  ;Set "seek" command code.
RqSK3:	call	DoCmd		;Issue desired command to drive.
	jc s	RqPLE		;If error, post return code & exit.
RqSKX:	ret			;Exit.
;
; CD/DVD "Play Audio" request handler.
;
RqPlay:	cmp	es:[si.RLSC].dwd,0  ;Is sector count zero?
	je s	RqSKX		    ;Yes, just exit above.
	mov	eax,es:[si+RLAddr]  ;Validate audio-start address.
	call	ValSN1
	mov	di,[bx+AudAP-@]	;Save drive's audio-start address.
	mov	[di],eax
	add	eax,es:[si+18]	;Calculate audio-end address.
	mov	edx,RMAXLBA	;Get maximum audio address.
	jc s	RqPL1		;If "end" WAY too big, use max.
	cmp	eax,edx		;Is "end" address past maximum?
	jbe s	RqPL2		;No, use "end" address as-is.
RqPL1:	xchg	eax,edx		;Set "end" address to maximum.
RqPL2:	mov	[di+4],eax	;Save drive's audio-end address.
	call	PlayA		;Issue "Play Audio" command.
RqPLE:	jc s	RqDS0		;If error, post return code & exit.
RqPLX:	jmp	RdAST4		;Go post "busy" status and exit.
;
; CD/DVD "Stop Audio" request handler.
;
RqStop:	mov	[bx+Packet-@].lb,04Bh ;Set "Pause/Resume" command.
	jmp	DoCmd		      ;Go pause "audio" and exit.
;
; CD/DVD "Resume Audio" request handler.
;
RqRsum:	inc	[bx+PktLn+1-@].lb  ;Set "Resume" flag for above.
	call	RqStop		;Issue "Pause/Resume" command.
	jmp s	RqPLE		;Go exit through "RqPlay" above.
;
; CD/DVD IOCTL Input "Device Header Address" handler.
;
ReqDHA:	push	ds		;Return our base driver address.
	push	bx
	pop	es:[si+1].dwd
	ret			;Exit.
;
; CD/DVD IOCTL Input "Current Head Location" handler.
;
RqHLoc:	mov	[bx+Packet-@].dwd,001400042h   ;Set command bytes.
	mov	al,16		;Set input byte count of 16.
	call	RdAST3		;Issue "Read Subchannel" request.
	jc s	RqDS0		;If error, post return code & exit.
	mov	es:[si+1],bl	;Return "HSG" addressing mode.
	call	SwpLBA		;Return "swapped" head location.
	mov	es:[si+2],eax
	jmp s	RqVSX		;Go post "busy" status and exit.
;
; CD/DVD IOCTL Input "Device Status" handler.
;
ReqDS:	mov	[bx+Packet-@].dwd,0002A005Ah  ;Set up mode-sense.
	mov	al,16		;Use input byte count of 16.
	call	DoBuf1		;Issue mode-sense for hardware data.
RqDS0:	jc s	RqVSE		;If error, post return code & exit.
	mov	eax,00214h	;Get our basic driver status flags.
	mov	cl,070h		;Get media status byte.
	xor	cl,[di+2]
	shr	cl,1		;Door open, or closed with no disk?
	jnz s	RqDS1		;No, check "drive locked" status.
	adc	ax,00800h	;Post "door open" & "no disk" flags.
RqDS1:	test	[di+14].lb,002h	;Drive pushbutton "locked out"?
	jnz s	RqVSS		;No, set flags in IOCTL and exit.
	or	al,002h		;Set "door locked" status flag.
	jmp s	RqVSS		;Go set flags in IOCTL and exit.
;
; CD/DVD IOCTL Input "Sector Size" handler.
;
ReqSS:	cmp	es:[si+1].lb,1	;Is read mode "cooked" or "raw"?
	ja	RqGenE		;Neither -- Post "general error".
	mov	ax,RAWSL	;Get "raw" sector length.
	je s	RqSS1		;If "raw" mode, set sector length.
	mov	ax,COOKSL	;Get "cooked" sector length.
RqSS1:	mov	es:[si+2],ax	;Post sector length in IOCTL packet.
	ret			;Exit.
;
; CD/DVD IOCTL Input "Volume Size" handler.
;
ReqVS:	mov	[bx+Packet-@].lb,025h  ;Set "Read Capacity" code.
	mov	al,008h		;Get 8 byte data-transfer length.
	call	DoBuf2		;Issue "Read Capacity" command.
RqVSE:	jc s	RqASIE		;If error, post return code & exit.
	mov	eax,[di]	;Get and "swap" volume size.
	call	Swp32
RqVSS:	mov	es:[si+1],eax	;Set desired IOCTL data value.
RqVSX:	jmp s	RqATIX		;Go post "busy" status and exit.
;
; CD/DVD IOCTL Input "Media-Change Status" handler.
;
ReqMCS:	call	DoCmd		;Issue "Test Unit Ready" command.
	mov	di,[bx+AudAP-@]	;Get media-change flag from table.
	mov	al,[di-1]
	mov	es:[si+1],al	;Return media-change flag to user.
	ret			;Exit.
;
; CD/DVD IOCTL Input "Audio Disk Info" handler.
;
ReqADI:	mov	al,0AAh		;Specify "lead-out" session number.
	call	RdTOC		;Read disk table-of-contents (TOC).
	mov	es:[si+3],eax	;Set "lead out" LBA addr. in IOCTL.
	mov	ax,[di+2]	;Set first & last tracks in IOCTL.
	mov	es:[si+1],ax
	jmp s	RqATIX		;Go post "busy" status and exit.
;
; CD/DVD IOCTL Input "Audio Track Info" handler.
;
ReqATI:	mov	al,es:[si+1]	;Specify desired session (track) no.
	call	RdTOC		;Read disk table-of-contents (TOC).
	mov	es:[si+2],eax	;Set track LBA address in IOCTL.
	mov	al,[di+5]
	shl	al,4
	mov	es:[si+6],al
RqATIX:	jmp	RdAST		;Go post "busy" status and exit.
;
; CD/DVD IOCTL Input "Audio Q-Channel Info" handler.
;
ReqAQI:	mov	ax,04010h	;Set "data in", use 16-byte count.
	call	RdAST2		;Read current "audio" status.
	jc s	RqASIE		;If error, post return code & exit.
	mov	eax,[di+5]	;Set ctrl/track/index in IOCTL.
	mov	es:[si+1],eax
	mov	eax,[di+13]	;Set time-on-track in IOCTL.
	mov	es:[si+4],eax
	mov	edx,[di+9]	;Get time-on-disk & clear high
	shl	edx,8		;  order time-on-track in IOCTL.
	jmp s	RqASI4		;Go set value in IOCTL and exit.
;
; CD/DVD IOCTL Input "Audio Status Info" handler.
;
ReqASI:	mov	ax,04010h	;Set "data in", use 16-byte count.
	call	RdAST2		;Read current "audio" status.
RqASIE:	jc s	RqErrJ		;If error, post return code & exit.
	mov	es:[si+1],bx	;Reset audio "paused" flag.
	xor	eax,eax		;Reset starting audio address.
	xor	edx,edx		;Reset ending audio address.
	cmp	[di+1].lb,011h  ;Is drive now "playing" audio?
	jne s	RqASI1		;No, check for audio "pause".
	mov	di,[bx+AudAP-@]	;Point to drive's audio data.
	mov	eax,[di]	;Get current audio "start" addr.
	jmp s	RqASI2		;Go get current audio "end" addr.
RqASI1:	cmp	[di+1].lb,012h	;Is drive now in audio "pause"?
	jne s	RqASI3		;No, return "null" addresses.
	inc	es:[si+1].lb	;Set audio "paused" flag.
	call	SwpLBA		;Convert time-on-disk to LBA addr.
	call	CvtLBA
	mov	di,[bx+AudAP-@]	;Point to drive's audio data.
RqASI2:	mov	edx,[di+4]	;Get current audio "end" address.
RqASI3:	mov	es:[si+3],eax	;Set audio "start" addr. in IOCTL.
RqASI4:	mov	es:[si+7],edx	;Set audio "end" address in IOCTL.
	ret			;Exit.
;
; CD/DVD IOCTL Output "Eject Disk" handler.
;
ReqEJ:	mov	[bx+Packet-@].lw,0011Bh  ;Set "eject" commands.
	mov	[bx+PktLBA+2-@].lb,002h  ;Set "eject" function.
	jmp	RqSK3			 ;Go do "eject" & exit.
;
; CD/DVD IOCTL Output "Lock/Unlock Door" handler.
;
ReqLU:	mov	al,es:[si+1]	;Get "lock" or "unlock" function.
	cmp	al,001h		;Is function byte too big?
	ja s	RqGenE		;Yes, post "general error" & exit.
	mov	cx,0001Eh	    ;Get "lock" & "unlock" commands.
RqLU1:	mov	[bx+Packet-@],cx    ;Set "packet" command bytes.
	mov	[bx+PktLBA+2-@],al  ;Set "packet" function byte.
	call	DoCmd		    ;Issue desired command to drive.
	jc s	RqErrJ		;If error, post return code & exit.
	jmp s	RdAST		;Go post "busy" status and exit.
;
; CD/DVD IOCTL Output "Reset Drive" handler.
;
ReqRS:	call	HltDMA		;Stop previous DMA & select drive.
	inc	dx		;Point to IDE command register.
	mov	al,008h		;Do an ATAPI "soft reset" command.
	out	dx,al
	call	ChkTO		;Await controller-ready.
	jnc s	SwpLBX		;If no timeout, go exit below.
RqGenE:	mov	al,GENERR	;General error!  Get return code.
RqErrJ:	jmp	ReqErr		;Go post error return code and exit.
;
; CD/DVD IOCTL Output "Close Tray" handler.
;
ReqCL:	mov	al,003h		;Get "close tray" function byte.
	mov	cx,0011Bh	;Get "eject" & "close" commands.
	jmp s	RqLU1		;Go do "close tray" command above.
;
; CD/DVD subroutine to read disk "Table of Contents" (TOC) values.
;
RdTOCE:	pop	cx		;Error -- Discard exit address.
	jmp s	RqErrJ		;Go post return code and exit.
RdTOC:	mov	[bx+Packet-@].lw,00243h  ;Set TOC and MSF bytes.
	call	DoTOC1		;Issue "Read Table of Contents" cmd.
	jc s	RdTOCE		;If error, exit immediately.
;
; CD/DVD subroutine to "swap" the 4 bytes of a a 32-bit value.
;
SwpLBA:	mov	eax,[di+8]	;Get audio-end or buffer LBA value.
Swp32:	xchg	al,ah		;"Swap" original low-order bytes.
	rol	eax,16		;"Exchange" low- and high-order.
	xchg	al,ah		;"Swap" ending low-order bytes.
SwpLBX:	ret			;Exit.
;
; CD/DVD subroutine to read current "audio" status & disk addresses.
;
RdAST:	call	ClrPkt		  ;Status only -- reset ATAPI packet.
RdAST1:	mov	ax,00004h	  ;Clear "data in", use 4-byte count.
RdAST2:	mov	[bx+Packet-@].dwd,001000242h  ;Set command bytes.
	mov	[bx+PktLBA-@],ah  ;Set "data in" flag (RdAST2 only).
RdAST3:	call	DoBuf1		  ;Issue "Read Subchannel" command.
	jc s	RdASTX		  ;If error, exit immediately.
	cmp	[di+1].lb,011h	  ;Is a "play audio" in progress?
	clc			  ;(Zero carry flag in any case).
	jne s	RdASTX		  ;No, just exit below.
RdAST4:	push	si		  ;Save SI- and ES-regs.
	push	es
	les	si,[bx+RqPkt-@]	  ;Reload DOS request-packet addr.
	or	es:[si.RPStat],RPBUSY  ;Set "busy" status bit.
	pop	es		  ;Reload ES- and SI-regs.
	pop	si
RdASTX:	ret			  ;Exit.
;
; Subroutine to handle CD/DVD Multi-Session disks for reads & seeks.
;   Multi-Session disks require (A) saving the last-session starting
;   LBA for a new disk after any media-change and (B) "offsetting" a
;   read of the VTOC or initial directory block, sector 16 or 17, to
;   access the VTOC/directory of the disk's last session.
;
MultiS:	mov	di,[bx+AudAP-@]	;Point to drive variables.
	cmp	[di+11].lb,0FFh	;Is last-session LBA valid?
	jne s	MultS1			;Yes, proceed with request.
	mov	[bx+Packet-@].lb,043h	;Set "Read TOC" command.
	inc	[bx+PktLBA-@].lb	;Set "format 1" request.
	call	DoTOC2			;Read first & last session.
	jc s	MultSX			;If any error, exit below.
	mov	[bx+PktLBA-@],bl	;Reset "format 1" request.
	mov	al,[di+3]		;Get last-session number.
	call	DoTOC1		;Read disk info for last session.
	jc s	MultSX		;If error, exit with carry set.
	call	SwpLBA		;"Swap" & save last-session LBA addr.
	mov	di,[bx+AudAP-@]
	mov	[di+8],eax
	call	ClrPkt		   ;Reset our ATAPI packet area.
MultS1:	mov	eax,es:[si+RLSec]  ;Get starting sector number.
	mov	dx,ax		   ;Save low-order sector number.
	and	al,0FEh		;"Mask" sector to an even number.
	cmp	eax,16		;Sector 16 (VTOC) or 17 (directory)?
	xchg	ax,dx		;(Restore low-order sector number).
	jne s	MultS2		;No, set sector in packet.
	add	eax,[di+8]	;Offset sector to last-session start.
MultS2:	call	Swp32		;"Swap" sector into packet as LBA.
	mov	[bx+PktLBA-@],eax
	clc			;Clear carry flag (no errors).
MultSX:	ret			;Exit.
;
; Subroutine to ensure CD/DVD UltraDMA is stopped and then select our
;   drive.   For some older chipsets, if UltraDMA is running, reading
;   an IDE register causes the chipset to "HANG"!!
;
HltDMA:	mov	dx,[bx+DMAAd-@]	;Get drive UltraDMA command address.
	test	dl,001h		;Is this drive using UltraDMA?
	jnz s	HltDM1		;No, just select "master" or "slave".
	in	al,dx		;Ensure any previous DMA is stopped!
	and	al,0FEh
	out	dx,al
HltDM1:	mov	dx,[bx+IdeDA-@]	;Point to IDE device-select register.
	add	dx,6
	mov	di,[bx+AudAP-@]	;Get drive's IDE device-select byte.
	mov	al,[di-2]
	out	dx,al		;Select IDE "master" or "slave" unit.
	ret			;Exit.
;
; Subroutine to read all data for a CD/DVD "Read Long" request.   This
;   is "called back" by our caching routine to read data NOT in cache!
;
RLRead:	sti			  ;Make all critical driver settings!
	cld
	push	08000h
@RLRSeg	equ	[$-2].lw	  ;(Driver segment address, Init set).
	pop	ds
	xor	bx,bx		  ;Zero BX-reg. for relative logic.
	push	cs		  ;Can current cache buffer be used?
	call	UseCB1		  ;Can current cache buffer be used?
	cmc			  ;If so, set XMS-move flag for below.
	adc	[bx+XMF-@],bl
	db	09Ah		  ;Move DMA command-list up to XMS.
	dw	(MovCLX-@)
@RLLSeg	dw	0		  ;(Driver segment address, Init set).
	jc s	RLRdE		  ;Failed?  Return "general error"!
	call	DoIO		  ;Do desired UltraDMA input request.
	jc s	RLRdX		  ;If any errors, exit below.
	or	[bx+XMF-@],bl	  ;Do we need an XMS buffer "move"?
	jz s	RLRdX		  ;No, just exit below.
	mov	ecx,[bx+VDSLn-@]  ;Get parameters for an XMS move.
	mov	esi,[bx+IOAdr-@]
	mov	edi,[bx+VDSAd-@]
	db	09Ah		  ;Move data from XMS to user buffer.
	dw	(MvData-@)
@RLMSeg	dw	0		  ;(Driver segment address, Init set).
RLRdE:	mov	al,GENERR	  ;Use "general error" code if error.
RLRdX:	retf			  ;Exit.
;
; Subroutine to issue a CD/DVD "Play Audio" command.   At entry, the
;   DI-reg. points to the audio-start address for this drive.
;
PlayA:	mov	eax,[di]	;Set "packet" audio-start address.
	call	CvtMSF
	mov	[bx+PktLBA+1-@],eax
	mov	eax,[di+4]	;Set "packet" audio-end address.
	call	CvtMSF
	mov	[bx+PktLH-@],eax
	mov	[bx+Packet-@].lb,047h  ;Set "Play Audio" command.
	jmp s	DoCmd		       ;Start playing audio & exit.
;
; Ye Olde CD/DVD I-O Subroutine.   All CD/DVD I-O gets executed here!
;
DoTOC1:	mov	[bx+PktLH-@],al	;"TOC" -- set session no. in packet.
DoTOC2:	mov	al,12		;Use 12-byte "TOC" allocation count.
DoBuf1:	mov	[bx+PktLn+1-@],al  ;Buffered -- set packet count.
DoBuf2:	xor	ah,ah		   ;Save data-transfer length.
	mov	[bx+VDSLn-@],ax
	mov	[bx+VDSOf-@].lw,(InBuf-@)  ;Use our buffer for I-O.
	mov	[bx+UserB+2-@],ds
	jmp s	DoIO		;Go start I-O below.
DoCmd:	mov	[bx+VDSLn-@],bx	;Command only -- reset xfr length.
DoIO:	push	si		;Save SI- and ES-registers.
	push	es
	mov	[bx+Try-@].lb,4	;Set request retry count of 4.
DoIO1:	call	HltDMA		;Stop previous DMA & select drive.
	call	ChkTO		;Await controller-ready.
	jc s	DoIO4		;Timeout!  Handle as a "hard error".
	cmp	[bx+DMF-@],bl	;UltraDMA input request?
	je s	DoIO2		;No, output ATAPI "packet" command.
	mov	al,008h		;Get DMA-input command bit.
	push	cs		;Initialize our DMA input transfer.
	call	IniDMA
DoIO2:	mov	ax,[bx+VDSOf-@]	;Reset data-transfer buffer address.
	mov	[bx+UserB-@],ax
	mov	ax,[bx+VDSLn-@]	;Reset data-transfer byte count.
	mov	[bx+UserL-@],ax
	mov	dx,[bx+IdeDA-@]	;Point to IDE "features" register.
	inc	dx
	mov	al,[bx+DMF-@]	;If UltraDMA input, set "DMA" flag.
	out	dx,al
	add	dx,3		;Point to byte count registers.
	mov	si,(UserL-@)	;Output data-transfer length.
	outsb
	inc	dx
	outsb
	inc	dx		;Point to command register.
	inc	dx
	mov	al,0A0h		;Issue "Packet" command.
	out	dx,al
	mov	cl,DRQ		;Await controller- and data-ready.
	call	ChkTO1
	jc s	DoIO4		;Timeout!  Handle as a "hard error".
	xchg	ax,si		;Save BIOS timer address.
	mov	dx,[bx+IdeDA-@]	;Point to IDE data register.
	mov	cl,6		;Output all 12 "Packet" bytes.
	mov	si,(Packet-@)
	rep	outsw
	xchg	ax,si		;Reload BIOS timer address.
	mov	ah,STARTTO	;Allow 7 seconds for drive startup.
	cmp	[bx+DMF-@],bl	;UltraDMA input request?
	je s	DoIO12		;No, start "PIO mode" transfer below.
	mov	[bx+UserL-@],bx	;Reset transfer length (DMA does it).
	mov	ch,ah		;Set 7-second timeout in AH-reg.
	add	ch,es:[si]
	push	cs		;Start and monitor next DMA transfer.
	call	RunDMA
	jnz s	DoIO4		;Problems?  Handle as a "hard error"!
	call	ChkTO		;Await final controller-ready.
	jc s	DoIO4		;Timeout!  Handle as a "hard error"!
DoIO3:	mov	si,[bx+AudAP-@]	;Get drive media-change flag pointer.
	dec	si
	and	ax,00001h	;Did controller detect any errors?
	jz s	DoIO6		;No, see if all data was transferred.
	sub	dx,6		;Get controller's sense key value.
	in	al,dx
	shr	al,4
	cmp	al,006h		;Is sense key "Unit Attention"?
	je s	DoIO7		;Yes, check for prior media-change.
	mov	ah,0FFh		;Get 0FFh flag for "Not Ready".
	cmp	al,002h		;Is sense key "Drive Not Ready"?
	je s	DoIO8		;Yes, go set our media-change flag.
DoIO4:	mov	dx,[bx+IdeDA-@]	;Hard error!  Point to command reg.
	add	dx,7
	mov	al,008h		;Issue ATAPI "soft reset" to drive.
	out	dx,al
	mov	al,11		;Get "hard error" return code.
DoIO5:	dec	[bx+Try-@].lb	;Do we have more I-O retries left?
	jz s	DoIO9		;No, set carry & return error code.
	jmp	DoIO1		;Try re-executing this I-O request.
DoIO6:	cmp	[bx+UserL-@],bx	;Was all desired data input?
	jne s	DoIO4		;No?  Handle as a "hard error".
	mov	[si].lb,001h	;Set "no media change" flag.
	jmp s	DoIO10		;Go reload regs. and exit below.
DoIO7:	mov	al,002h		;"Attention":  Get "Not Ready" code.
	cmp	[si],bl		;Is media-change flag already set?
	jle s	DoIO5		;Yes, retry & see if it goes away!
DoIO8:	xchg	ah,[si]		;Load & set our media-change flag.
	or	[bx+IOF-@].lb,1	;Flush cache on next caching request.
	mov	[si+12].lb,0FFh	;Make last-session LBA invalid.
	dec	ah		;Is media-change flag already set?
	jnz s	DoIO9		;Yes, set carry flag and exit.
	mov	al,15		;Return "Invalid Media Change".
DoIO9:	stc			;Set carry flag (error!).
DoIO10:	pop	es		;Reload ES- and SI-registers.
	pop	si
	mov	di,(InBuf-@)	;For audio, point to our buffer.
	ret			;Exit.
DoIO11:	mov	ah,SEEKTO	;"PIO mode" -- Get "seek" timeout.
DoIO12:	call	ChkTO2		;Await controller-ready.
	jc s	DoIO4		;Timeout!  Handle as a "hard error".
	test	al,DRQ		;Did we also get a data-request?
	jz s	DoIO3		;No, go check for any input errors.
	dec	dx		;Get controller-buffer byte count.
	dec	dx
	in	al,dx
	mov	ah,al
	dec	dx
	in	al,dx
	mov	dx,[bx+IdeDA-@]	;Point to IDE data register.
	mov	si,[bx+UserL-@]	;Get our data-transfer length.
	or	si,si		;Any remaining bytes to input?
	jz s	DoIO14		;No, "eat" all residual data.
	cmp	si,ax		;Remaining bytes > buffer count?
	jbe s	DoIO13		;No, input all remaining bytes.
	mov	si,ax		;Use buffer count as input count.
DoIO13:	les	di,[bx+UserB-@]	;Get input data-transfer address.
	mov	cx,si		;Input all 16-bit data words.
	shr	cx,1
	rep	insw
	add	[bx+UserB-@],si	;Increment data-transfer address.
	sub	[bx+UserL-@],si	;Decrement data-transfer length.
	sub	ax,si		;Any data left in controller buffer?
	jz s	DoIO11		;No, await next controller-ready.
DoIO14:	xchg	ax,cx		;"Eat" all residual input data.
	shr	cx,1		;(Should be NO residual data as we
DoIO15:	in	ax,dx		;  always set an exact byte count.
	loop	DoIO15		;  This logic is only to be SAFE!).
	jmp s	DoIO11		;Go await next controller-ready.
	db	0		;(Unused alignment "filler").
;
; Subroutine to test for CD/DVD I-O timeouts.   At entry the CL-reg.
;   is 008h to test for a data-request, also.   At exit, the DX-reg.
;   points to the drive's IDE primary-status reg.   The AH-, SI- and
;   ES-regs. will be lost.
;
ChkTO:	xor	cx,cx		;Check for only controller-ready.
ChkTO1:	mov	ah,CMDTO	;Use 500-msec command timeout.
ChkTO2:	mov	es,bx		;Point to low-memory BIOS timer.
	mov	si,BIOSTMR
	add	ah,es:[si]	;Set timeout limit in AH-reg.
ChkTO3:	cmp	ah,es:[si]	;Has our I-O timed out?
	stc			;(If so, set carry flag).
	je s	ChkTOX		;Yes?  Exit with carry flag on.
	mov	dx,[bx+IdeDA-@]	;Read IDE primary status.
	add	dx,7
	in	al,dx
	test	al,BSY		;Is our controller still busy?
	jnz s	ChkTO3		;Yes, loop back and test again.
	or	cl,cl		;Are we also awaiting I-O data?
	jz s	ChkTOX		;No, just exit.
	test	al,cl		;Is data-request (DRQ) also set?
	jz s	ChkTO3		;No, loop back and test again.
ChkTOX:	ret			;Exit:  Carry "on" denotes a timeout!
;
; Subroutine to convert an LBA sector number to "RedBook" minutes,
;   seconds, and "frames" format (MSF).
;
CvtMSF:	add	eax,150		;Add in offset.
	push	eax		;Get address in DX:AX-regs.
	pop	ax
	pop	dx
	mov	cx,75		;Divide by 75 "frames"/second.
	div	cx
	shl	eax,16		;Set "frames" remainder in upper EAX.
	mov	al,dl
	ror	eax,16
	mov	cl,60		;Divide quotient by 60 seconds/min.
	div	cl
	ret			;Exit -- EAX-reg. contains MSF value.
;
; Subroutine to clear our ATAPI "packet" area for the next request.
;
ClrPkt:	mov	[bx+Packet-@],bx    ;Zero 1st 10 ATAPI packet bytes.
	mov	[bx+Packet+2-@],bx  ;(Last 2 are unused "pad" bytes).
	mov	[bx+Packet+4-@],bx
	mov	[bx+Packet+6-@],bx
	mov	[bx+Packet+8-@],bx
	ret			    ;Exit.
;
; Subroutine to set the 32-bit DMA address for an UltraDMA transfer.
;   At entry the EAX-reg. has the address from a VDS "lock" request.
;   At exit, if this address was unsuitable for UltraDMA, the driver
;   "Main XMS buffer" address has been set and the carry flag is on.
;
SetAdr:	test	al,003h		      ;Is user buffer 32-bit aligned?
	jnz s	SetAd1		      ;No, use our XMS memory buffer.
	cmp	[bx+VDSAd-@].hw,009h  ;Is DMA beyond our 640K limit?
	ja s	SetAd1		      ;Yes, use our XMS memory buffer.
	mov	cx,[bx+IOLen-@]	  ;Get (byte count - 1) + I-O address.
	dec	cx
	add	cx,ax		  ;Would I-O cross a 64K boundary?
	jnc s	SetAd2		  ;No, set final buffer address.
SetAd1:	stc			  ;Set carry to denote "our buffer".
	mov	eax,0		  ;Get our 32-bit XMS buffer addr.
@XBAddr	equ	[$-4].dwd	  ;(XMS buffer address, Init set).
SetAd2:	mov	[bx+IOAdr-@],eax  ;Set final 32-bit buffer addr.
	retf			  ;Exit.
;
; Subroutine to convert "RedBook" MSF values to an LBA sector number.
;
CvtLBA:	mov	cx,ax		;Save "seconds" & "frames" in CX-reg.
	shr	eax,16		;Get "minute" value.
	cmp	ax,99		;Is "minute" value too large?
	ja s	CvtLBE		;Yes, return -1 error value.
	cmp	ch,60		;Is "second" value too large?
	ja s	CvtLBE		;Yes, return -1 error value.
	cmp	cl,75		;Is "frame" value too large?
	ja s	CvtLBE		;Yes, return -1 error value.
	xor	edx,edx		;Zero EDX-reg. for 32-bit math below.
	mov	dl,60		;Convert "minute" value to "seconds".
	mul	dl		;(Multiply by 60, obviously!).
	mov	dl,ch		;Add in "second" value.
	add	ax,dx
	mov	dl,75		;Convert "second" value to "frames".
	mul	edx		;(Multiply by 75 "frames"/second).
	mov	dl,150		;Subtract offset - "frame".
	sub	dl,cl		;("Adds" frame, "subtracts" offset).
	sub	eax,edx
	ret			;Exit.
CvtLBE:	or	eax,-1		;Too large!  Set -1 error value.
	ret			;Exit.
;
; Subroutine to test if the current cache buffer can be used for DMA.
;   If so, the cache-buffer address is set for UltraDMA and the carry
;   flag is "off" at exit.   If not, or if the stand-alone driver (no
;   cache) is in use, the carry flag is "on" at exit.    The logic at
;   "UseCB" is here, not in "UdmaIO", to reduce the 250-MB HMA size.
;
UseCB:	mov	[bx+SecCt-@],al	      ;Set disk I-O sector count.
	mov	ah,0		      ;Set disk DMA buffer length.
	shl	ax,1
	mov	[bx+IOLen+1-@],ax
	mov	ax,[bx+RqLBA+4-@]     ;Set LBA bits 0 thru 15.
	mov	[bx+LBA-@],ax
	mov	eax,[bx+RqBuf-@]      ;Get user I-O buffer address.
UseCB1:	bt	[bx+DMAAd-@],bx	      ;CD/DVD drive without UltraDMA?
				      ;("stc"/"retf", if no caching).
	jc s	UseCBX		      ;Yes, exit and use regular I-O.
	mov	cl,[bx+RqCSC-@]	      ;Get current I-O sector count.
	cmp	cl,[bx+RqSec-@]	      ;Is all I-O in one cache block?
	jb s	UseCBX		      ;No, exit and use regular I-O.
	or	[bx+RqCmd-@].lb,080h  ;Do NOT cache data, after I-O!
	mov	eax,[bx+RqXMS-@]      ;Do DMA thru the cache buffer!
	mov	[bx+IOAdr-@],eax
UseCBX:	retf			      ;Exit.
	db	0		      ;(Unused alignment "filler").
;
; Subroutine to start and monitor the next UltraDMA I-O transfer.
;
RunDMA:	mov	es:[si+HDI_OFS],bl  ;Reset BIOS disk-interrupt flag.
	mov	dx,[bx+DMAAd-@]	    ;Point to DMA command register.
	in	al,dx		;Set DMA Start/Stop bit (starts DMA).
	inc	ax
	out	dx,al
RunDM1:	inc	dx		;Point to DMA status register.
	inc	dx
	in	ax,dx		;Read DMA controller status.
	dec	dx		;Point back to DMA command register.
	dec	dx
	and	al,DMI+DME	;DMA interrupt or DMA error?
	jnz s	RunDM2		;Yes, halt DMA and check results.
	cmp	ch,es:[si]	;Has our DMA transfer timed out?
	je s	RunDM2		    ;Yes?  Halt DMA & post timeout!
	cmp	es:[si+HDI_OFS],bl  ;Did BIOS get a disk interrupt?
	je s	RunDM1		    ;No, loop back & retest status.
	mov	al,DMI		;Set "simulated" interrupt flag.
RunDM2:	push	ax		;Save ending DMA status.
	in	al,dx		;Reset DMA Start/Stop bit.
	and	al,0FEh
	out	dx,al
	pop	ax		;Reload ending DMA status.
	cmp	al,DMI		;Did DMA end with only an interrupt?
	jne s	RunDMX		;No?  Exit with "non-zero" (error!).
	inc	dx		;Reread DMA controller status.
	inc	dx
	in	al,dx
	test	al,DME		;Any "late" DMA error after DMA end?
RunDMX:	retf			;Exit -- "Non-zero" indicates error.
NormSz2	equ	$+STACK		;Size of "large" normal-memory logic.
				;  All logic below may go in the HMA.
;
; Ye Olde UltraDMA "Internal Driver" Subroutine.   This routine sets
;   up I-O requests, then calls "DoDMA" below to perform actual I-O.
;
UdmaIO:	cld			;Ensure FORWARD "string" commands.
	mov	bp,[bx+RqTyp-@]	;Get this disk's "type" flags.
	mov	ax,bp		;Get disk's IDE controller number.
	shr	al,2
	mov	ah,CTLTSIZ	;Point to disk's I-O addresses.
	mul	ah
	add	ax,(Ctl1Tbl-@)
	xchg	ax,si
	lodsd			   ;Set controller base address and
	mov	[bx+DMAAd-@],eax   ;  primary-channel data address.
	rcr	bp,2		   ;Primary channel I-O request?
	jnc s	GetSec		   ;Yes, get sector count & commands.
	add	[bx+DMAAd-@].lw,8  ;Use secondary DMA controller.
	lodsw			   ;Set secondary channel data addr.
	mov	[bx+IdeDA-@],ax
GetSec:	mov	ax,[bx+RqSec-@]	   ;Get sector count and command bits.
	les	di,[bx+RqLBA-@]	   ;Set LBA bits 16 thru 47.
	mov	dx,es
	mov	[bx+LBA2+1-@],di
	mov	[bx+LBA2-@],dh
	mov	[bx+LBA+2-@],dl
	shr	dx,12		   ;Shift out LBA bits 16-27.
	or	di,dx		   ;Anything in LBA bits 28-47?
	jnz s	LBA48		   ;Yes, use LBA48 read/write command.
	xchg	dh,[bx+LBA2-@]	   ;LBA28:  Reload & reset bits 24-27.
	or	ah,(DRCMD+1)	   ;Get LBA28 read/write command + 5.
	jmp s	SetCmd		   ;Go set LBA and IDE command bytes.
LBA48:	shl	ah,3		   ;LBA48 -- get command as 020h/030h.
SetCmd:	shl	bp,1		   ;Shift master/slave into LBA cmds.
	mov	dl,(LBABITS/32)
	rcl	dl,5
	or	dl,dh		   ;"Or" in LBA28 bits 24-27 (if any).
	mov	dh,005h		   ;Get final IDE command byte.
	xor	dh,ah		   ;(LBA28 C8h/CAh, LBA48 25h/35h).
	mov	[bx+DSCmd-@],dx	   ;Set LBA and IDE command bytes.
	push	cs		   ;Can current cache buffer be used?
@UCBAd:	call	UseCB
	nop			   ;(Unused "filler" for 3.5K driver).
	jnc s	UBufRq		   ;Yes, go do buffered I-O below.
	push	cs		   ;Set user or Main XMS buffer addr.
@USAAd:	call	SetAdr
	nop			   ;(Unused "filler" for 3.5K driver).
	jnc s	DoDMA		     ;If user buffer, do direct DMA.
UBufRq:	test	[bx+RqCmd-@].lb,002h ;Buffered I-O:  Read request?
	jz s	UBufIn		     ;Yes, use input routine below.
	push	cs		     ;Move user data to XMS buffer.
	call	UBufMv
	jnc s	DoDMA		     ;If no error, do output & exit.
UBufEr:	retf			     ;XMS or I-O error:  Exit now!
UBufIn:	push	cs		     ;Buffered input:  Call "DoDMA"
	call	DoDMA		     ;  and input to our XMS buffer.
	jc s	UBufEr		     ;If any I-O errors, exit above!
UBufMv:	movzx	ecx,[bx+SecCt-@].lb  ;Get number of sectors to move.
	mov	esi,[bx+IOAdr-@]     ;Set move addresses for a read.
	mov	edi,esi
	xchg	edi,[bx+RqBuf-@]
;
; Subroutine to do XMS moves for logic that may reside in the HMA.
;
HMAMov:	jz s	HMAMv1		;Is this a read from XMS memory?
	xchg	esi,edi		;No, "swap" source & destination.
HMAMv1:	shl	ecx,9		;Convert sectors to byte count.
	db	0EAh		;Go move all XMS data, then exit.
	dw	(MvData-@)
@HMVSeg	dw	0		;(Driver segment address, Init set).
;
; Subroutine to handle hard-disk UltraDMA I-O requests for "UdmaIO".
;
DoDMA:	db	09Ah		;Move DMA command-list up to XMS.
	dw	(MovCLX-@)
@DDLSeg	dw	0		;(Driver segment address, Init set).
	jc s	DoDMAX		;Command-list move ERROR?  Exit NOW!
	mov	dx,[bx+DMAAd-@]	;Ensure any previous DMA is stopped!
	in	al,dx		;(On some older chipsets, if DMA is
	and	al,0FEh		;  running, reading an IDE register
	out	dx,al		;  causes the chipset to "HANG"!!).
	mov	di,[bx+IdeDA-@]	;Get our disk's IDE base address.
	mov	al,[bx+DSCmd-@]	;Select our desired disk.
	and	al,0F0h
	lea	dx,[di+CDSEL]
	out	dx,al
	mov	es,bx		;Point to low-memory BIOS timer.
	mov	si,BIOSTMR
	mov	cx,((RDYTO*256)+FLT)  ;Get timeout & "fault" mask.
	add	ch,es:[si]	      ;Set timeout limit in CH-reg.
	push	cs		      ;Await controller/disk ready.
	call	ChkRdy
	jc s	DoDMAX		      ;If any errors, exit below!
	test	[bx+IOCmd-@].lb,012h  ;Is this a write request?
	jnz s	SetDMA		      ;Yes, init DMA transfer.
	mov	al,008h		;Get "DMA read" command bit.
SetDMA:	push	si		;Save BIOS timer pointer.
	push	cs		;Initialize this DMA I-O transfer.
@DDIAd:	call	IniDMA
	nop			;(Unused "filler", for 3.5K driver).
	mov	ax,001F7h	;Set IDE parameter-output flags.
NxtPar:	lea	dx,[di+CDATA+1]	;Point to IDE sector count reg. - 1.
IDEPar:	inc	dx		;Output all ten LBA48 parameter bytes.
	outsb			;(1st 4 overlayed by 2nd 4 if LBA28!).
	shr	ax,1		;More parameters to go in this group?
	jc s	IDEPar		;Yes, loop back and output next one.
	jnz s	NxtPar		;If first 4 done, go output last 6.
	pop	si		;Reload BIOS timer pointer.
	dec	dh		;"Legacy IDE" controller channel?
@DRQ:	jmp s	DMAGo		;No, forget about awaiting 1st DRQ.
	mov	dh,003h		;Get IDE alternate-status address.
	dec	dx		;(Primary-status address | 300h - 1).
ChkDRQ:	cmp	ch,es:[si]	;Too long without 1st data-request?
	je s	ErrDMA		;Yes?  Return carry and DMA error!
	in	al,dx		;Read IDE alternate status.
	and	al,DRQ		;Has 1st data-request arrived?
	jz s	ChkDRQ		;No, loop back and check again.
	nop			;(Unused alignment "filler").
DMAGo:	push	cs		;Start and monitor our DMA transfer.
@DDRAd:	call	RunDMA
	nop			;(Unused "filler", for 3.5K driver).
	jnz s	ErrDMA		;Problems?  Return carry & DMA error!
	inc	cx		;Check "fault" and hard-error at end.
ChkRdy:	lea	dx,[di+CSTAT]	;Read IDE primary status.
	in	al,dx
	cmp	ch,es:[si]	;Too long without becoming ready?
	je s	RdyErr		;Yes?  Go see what went wrong.
	test	al,BSY+RDY	;Controller or disk still busy?
	jle s	ChkRdy		;Yes, loop back and check again.
	and	al,cl		;Disk "fault" or hard-error?
	jnz s	HdwErr		;Yes?  Go see what went wrong.
DoDMAX:	retf			;End of request -- Return to "UdmaIO".
ErrDMA:	mov	al,DMAERR	;BAAAD News!  Post DMA error code.
	stc			;Set carry flag (error!) and exit.
	retf
RdyErr:	test	al,BSY		;BAAAD News!  Did controller go ready?
	mov	ax,(256*CTLRERR)+DISKERR ;(Get not-ready error codes).
	jmp s	WhichE		;Go see which error code to return.
HdwErr:	test	al,FLT		;BAAAD News!  Is the disk "faulted"?
	mov	ax,(256*FAULTED)+HARDERR ;(Get hardware error codes).
WhichE:	jz s	Kaput		;If "zero", use AL-reg. return code.
	mov	al,ah		;Use AH-reg. return code of this pair.
Kaput:	stc			;Set carry flag (error!) and exit.
	retf
DvrEnd	equ	$+STACK		;Size of stand-alone driver.
;
; Main Caching Routine.
;
Cache:	btr	[bx+IOF-@],bx	;Does cache need to be "flushed"?
	jnc s	NextIO		;No, proceed with this I-O request.
	mov	ax,0FFFFh	;Get binary-search table data.
@TC1	equ	[$-2].lw	     ;(Table count - 1, Init set).
	mov	si,offset (SrchT-@)
@TE1	equ	[$-2].lw	     ;(Table upper-limit, Init set).
FlushT:	dec	si		     ;Flush our binary-search table.
	dec	si
	mov	cs:[si],ax
	dec	ax
	jns s	FlushT
	mov	[bx+STLmt-@],si	     ;Reset "free" cache-index ptr.
	or	[bx+LUTop-@].dwd,-1  ;Reset LRU start & end indexes.
NextIO:	push	ds		     ;Set ES-reg. same as DS-reg.
	pop	es
	mov	al,0		;Get "granularity" (sectors/block).
@GRAN1	equ	[$-1].lb	;(Block "granularity", set by Init).
	cmp	al,[bx+RqSec-@]	;Will we need multiple cache blocks?
	jbe s	SetSC		;Yes, use "full block" sector count.
	mov	al,[bx+RqSec-@]	;Use remaining request sector count.
SetSC:	mov	[bx+RqCSC-@],al	;Set maximum I-O sector count.
	mov	[bx+RqXMS-@],bx ;Reset lower XMS buffer offset.
	mov	dx,0		;Set up for our binary search.
@MP1	equ	[$-2].lw	;(Beginning offset, set by Init).
	mov	bp,offset (SrchT-2-@)
@MP2	equ	[$-2].lw	;(Midpoint address, set by Init).
Search:	shr	dx,1		;Divide binary-search offset by 2.
	and	dl,0FEh		;Ensure offset remains "even".
	cmp	bp,[bx+STLmt-@]	;Is our search pointer too high?
	jae s	SrchLo		;Yes, cut search pointer by offset.
	mov	ax,cs:[bp]	;Get next binary-search table index.
	call	CBGet		;Get next cache entry into buffer.
	jc	EndIO		;BAAAD News if XMS error -- go exit!
	mov	si,(RqLBA-@)	;Compare initial request & table LBA.
	call	CComp
	jae s	ChkEnd		;Request >= table:  Check for found.
SrchLo:	sub	bp,dx		;Subtract offset from search ptr.
	or	dx,dx		;Offset zero, i.e. last compare?
	jnz s	Search		;No, go compare next table entry.
	jmp s	NoFind		;Handle this request as "not found".
ChkEnd:	je s	Found		;Request = table:  Treat as "found".
	mov	cl,[bx+CBSec-@]	;Calculate and set ending entry LBA.
	mov	si,(CBLBA-@)
	call	CalcEA
	mov	si,(RqLBA-@)	;Compare request start & entry end.
	call	CComp1
	jb s	Found		   ;Request < Entry:  Handle as found.
	ja s	SrchHi		   ;Request > Entry:  Bump search ptr.
	cmp	[bx+CBSec-@].lb,0  ;Is this cache entry "full"?
@GRAN2	equ	[$-1].lb	   ;(Block "granularity", Init set).
	jb s	Found		;No, handle this request as "found".
SrchHi:	add	bp,dx		;Add offset to search pointer.
	or	dx,dx		;Offset zero, i.e. last compare?
	jnz s	Search		;No, go compare next table entry.
	inc	bp		;Bump pointer to next table entry.
	inc	bp
NoFind:	push	cs		;Unfound:  Set search table segment.
	pop	es
	mov	di,[bx+STLmt-@]	;Get next "free" cache-entry index.
	mov	ax,es:[di]
	std			;Enable a "backward" move, and set
	lea	si,[di-2]	;  move-up source address & count.
	mov	cx,di		;  We do the "std" early, for more
	sub	cx,bp		;  time margin v.s. AMD CPU "bugs"!
	shr	cx,1
	rep	movs es:[di].lw,cs:[si].lw  ;Move up higher indexes.
	stosw				    ;Store new search index.
	cld			   ;Re-enable FORWARD "string" cmds.
	push	ds		   ;Reload this driver's ES-reg.
	pop	es
	add	[bx+STLmt-@].lw,2  ;Bump binary-search table limit.
	mov	si,(RqLBA-@)	   ;Set 48-bit LBA in new entry.
	mov	di,(CBLBA-@)
	movsd
	movsw
	movsb			   ;Set "cache unit" in new entry.
	mov	[di],bl		   ;Reset new entry's sector count.
Found:	mov	RqIndex,ax	   ;Post current cache-entry index.
	mov	cx,[bx+RqLBA+4-@]  ;Get starting I-O offset in block.
	sub	cx,[bx+CBLBA+4-@]
	mov	[bx+RqXMS-@],cl	   ;Set starting XMS sector offset.
	mov	[bx+LBA2+1-@],cx   ;Save starting I-O sector offset.
	cmp	[bx+CBSec-@],bl	   ;Is this a new cache-table entry?
	je s	ReLink		   ;Yes, relink entry as top-of-list.
	push	ax		   ;Unlink this entry from LRU list.
	call	UnLink
	pop	ax
ReLink:	mov	[bx+RqXMS+2-@],bx  ;Reset upper XMS buffer offset.
	movzx	edx,ax		   ;Get 32-bit cache block number.
	shl	edx,2		   ;Shift number to starting sector.
@GRANSC	equ	[$-1].lb	   ;("Granularity" shift, Init set).
	add	[bx+RqXMS-@],edx   ;Add to "preset" sector offset.
	shl	[bx+RqXMS-@].dwd,9	  ;Convert sectors to bytes.
	add	[bx+RqXMS-@].dwd,020000h  ;Add in XMS "base" address.
@XBase	equ	[$-4].dwd		  ;(XMS "base", set by Init).
	mov	cx,0FFFFh	;Make this entry "top of list".
	or	[bx+CBLst-@],cx	;Set this entry's "last" index.
	mov	dx,ax		;Swap top-of-list & entry index.
	xchg	dx,[bx+LUTop-@]
	mov	[bx+CBNxt-@],dx ;Set this entry's "next" index.
	cmp	dx,cx		;Is this the only LRU index?
	mov	cx,ax		;(Get this entry's index in CX-reg.).
	je s	ReLnk1		;Yes, make entry last on LRU list.
	push	ax		;Link entry to prior "top of list".
	call	UnLnk3
	pop	ax
	jmp s	ReLnk2		;Go deal with I-O sector count.
Update:	mov	[bx+CBSec-@],cl	;Update this entry's total sectors.
	call	CBPut		;Update this cache-table entry.
	jc s	EndIOJ		;If any XMS error, exit fast!
	movzx	cx,[bx+RqCSC-@]	;Calculate ending LBA for this I-O.
	mov	si,(RqLBA-@)
	call	CalcEA
	inc	bp		;Point to next search index.
	inc	bp
Ovrlap:	cmp	bp,[bx+STLmt-@]	;More cache-table entries to check?
	jae s	CachIO		;No, O.K. to handle "found" entry.
	mov	ax,cs:[bp]	;Set next binary-search table index.
	call	CBGet		;Get next cache entry into buffer.
	jc s	EndIOJ		;BAAAD News if XMS error -- go exit!
	mov	si,(LBABuf-@)	;Compare request end & entry start.
	call	CComp
	jbe s	CachIO		;Request <= entry:  O.K. to proceed.
	call	Delete		;Delete this overlapping table entry.
	jmp s	Ovrlap		;Go check for more entry overlap.
ReLnk1:	mov	[bx+LUEnd-@],ax	  ;Make entry last on LRU list, too!
	nop			  ;(Unused alignment "filler").
ReLnk2:	mov	cx,[bx+LBA2+1-@]  ;Reload initial I-O sector offset.
	mov	ch,0		  ;Get entry's available sectors.
@GRAN3	equ	[$-1].lb	  ;(Block "granularity", Init set).
	sub	ch,cl
	cmp	ch,[bx+RqCSC-@]	;More I-O sectors than available?
	jae s	Larger		;No, retain maximum sector count.
	mov	[bx+RqCSC-@],ch	;Reduce current I-O sector count.
	nop			;(Unused alignment "filler").
Larger:	add	cl,[bx+RqCSC-@]	;Get ending I-O sector number.
	cmp	cl,[bx+CBSec-@]	;More sectors than entry has now?
	ja s	Update		;Yes, update entry sectors.
	inc	bx		;Reset Z-flag for "put" into XMS.
	call	CBPut		;Update this cache-table entry.
EndIOJ:	jc s	EndIO		      ;If any XMS error, exit fast!
	test	[bx+RqCmd-@].lb,002h  ;Is this a write request?
	jz s	BufMov		      ;Yes, move cache data to user.
CachIO:	bts	[bx+RqCmd-@].lb,6  ;I-O done during a prior block?
	jc s	BfMore		   ;Yes, buffer more data into cache.
	mov	si,0F000h	   ;Get this unit's device-type flag.
	and	si,[bx+RqTyp-1-@]  ;Does the BIOS do this unit's I-O?
	js s	BiosIO		   ;Yes, go "call the BIOS" below.
	shr	si,10		;Get "internal driver" table offset.
	call	DvrTbl[si].dwd	;Call this "internal driver" for I-O.
	jmp s	CachRS		;Go restore critical driver settings!
BiosIO:	lds	si,CStack	;BIOS I-O:  Reload caller CPU flags.
	push	[si+2].lw
	popf
	pop	es		;Reload all CPU registers.
	pop	ds
	popad
	pushf			;"Call the BIOS" to do this request.
	db	09Ah
@Prv13B	dd	0		;(Prior Int 13h vector, set by Init).
	pushad			;Save all CPU registers again.
	push	ds
	push	es
	mov	al,ah		;Move any BIOS error code to AL-reg.
CachRS:	sti			;Restore critical driver settings.
	cld
	push	08000h		;Reload this driver's DS- & ES-regs.
@CRSSeg	equ	[$-2].lw	;(Driver segment address, Init set.
	pop	ds		;   MUST reload this way, since we
	push	ds		;   could be running from the HMA!).
	pop	es
	mov	bx,0		;Rezero BX-reg. but save carry bit.
	jnc s	InCach		;If good I-O, see if data was cached.
IOErr:	push	ax		;BAAAD News!  Save posted error code.
	mov	ax,RqIndex	;Delete current entry from the cache.
	call	ScanD
	pop	ax		;Reload returned I-O error code.
	stc			;Set carry again to denote error.
	xchg	bx,bx		;(Unused alignment "filler").
EndIO:	db	0EAh		;Return to driver "exit" logic.
	dw	(HDDRtn-@)
@HDXSeg	dw	0		;(Driver segment address, Init set).
InCach:	or	[bx+RqCmd-@],bl	;Did user data get cached during I-O?
	js s	ChkFul		;Yes, check if cache tables are full.
BfMore:	or	al,1		     ;Ensure we LOAD data into cache!
BufMov:	movzx	ecx,[bx+RqCSC-@].lb  ;Set XMS move sector count.
	mov	esi,[bx+RqXMS-@]     ;Set desired XMS buffer address.
	mov	edi,[bx+RqBuf-@]   ;Set user buffer address as dest.
	push	cs		   ;Move data between user and XMS.
	call	HMAMov
	jc s	IOErr		   ;If move error, use routine above.
	nop				   ;(Alignment "filler").
ChkFul:	cmp	[bx+STLmt-@].lw,(SrchT-@)  ;Binary-search table full?
@TE2	equ	[$-2].lw		   ;(Table "end", Init set).
	jb s	MoreIO		   ;No, check for more sectors to go.
	mov	ax,[bx+LUEnd-@]	   ;Delete least-recently-used entry.
	call	ScanD
MoreIO:	xor	ax,ax		   ;Reset error code for exit above.
	movzx	cx,[bx+RqCSC-@]	   ;Get current I-O sector count.
	sub	[bx+RqSec-@],cl	   ;More data sectors left to handle?
	jz s	EndIO		   ;No, return to our "exit" routine.
	add	[bx+RqLBA+4-@],cx  ;Update current LBA for next I-O.
	adc	[bx+RqLBA+2-@],bx
	adc	[bx+RqLBA-@],bx
	shl	cx,1		   ;Convert sector count to bytes.
	add	[bx+RqBuf+1-@],cl  ;Update user I-O buffer address.
	adc	[bx+RqBuf+2-@],bx
	jmp	NextIO		   ;Go handle next cache data block.
;
; Subroutine to calculate ending request or cache-table LBA values.
;
CalcEA:	mov	di,(LBABuf-@)	;Point to address-comparison buffer.
	push	[si].dwd	;Move high-order LBA into buffer.
	pop	[di].dwd
	add	cx,[si+4]	;Calculate ending LBA.
	mov	[di+4],cx
	adc	[di+2],bx
	adc	[di],bx
CalcEX:	ret			;Exit.
;
; Subroutine to do 7-byte "cache unit" and LBA comparisons.
;
CComp:	mov	di,(CBLBA-@)	;Point to current-buffer LBA value.
CComp1:	mov	cl,[bx+RqUNo-@]	;Compare our "unit" & table "unit".
	cmp	cl,[bx+CBUNo-@]
	jne s	CalcEX		;If units are different, exit now.
	mov	cx,3		;Compare our LBA & cache-table LBA.
	rep	cmpsw
	ret			;Exit -- Main routine checks results.
;
; Subroutine to set up and execute cache-entry XMS moves.
;
WBGet:	xor	di,di		;Work buffer:  Set Z-flag for "get".
WBPut:	mov	edi,(WBLBA-@)	;Set 32-bit "working" buffer address.
@WBAdr	equ	[$-4].dwd	;(Working-buffer address, Init set).
	jmp s	CEMov		;Go move data to/from current buffer.
CBGet:	xor	di,di		;Entry buffer:  Set Z-flag for "get".
CBPut:	mov	edi,(CBLBA-@)	;Set 32-bit "current" buffer address.
@CBAdr	equ	[$-4].dwd	;(Current-buffer address, Init set).
CEMov:	push	ax		;Save all needed 16-bit regs.
	push	cx
	push	dx
	pushf			;Save CPU Z-flag from above.
	xor	ecx,ecx		;Set cache-table entry size.
	mov	cl,12
	mul	cx		;Multiply cache index by entry size.
	push	dx		;Get cache-entry offset in ESI-reg.
	push	ax
	pop	esi
	add	esi,010000h	;Add cache-table "base" address.
@CTBase	equ	[$-4].dwd	;(Cache-table "base", Init set).
	popf			;Reload CPU flags.
	jz s	CEMov1		;Are we "getting" a cache entry?
	xchg	esi,edi		;No, "swap" source and destination.
CEMov1:	push	bp		;Save BP-reg. now (helps alignment).
	db	09Ah		;"Get" or "put" desired cache entry.
	dw	(MvData-@)
@CEMSeg	dw	0		;(Driver segment address, Init set).
	pop	bp		;Reload all 16-bit regs. we used.
	pop	dx
	pop	cx
	pop	ax		;Reload cache-entry index.
	jnc s	CalcEX		;If no BIOS/EMM386 move error, exit.
	or	[bx+IOF-@].lb,1	;BAAAD News!  Post our "flush" flag.
	stc			;Set carry again after "or" command.
	mov	al,XMSERR	;Post "XMS error" return code.
	ret			;Exit.
;
; Subroutine to scan for a binary-search table index and delete it.
;
ScanD:	mov	cx,0		;Set up a 16-bit "scan" command.
@TC2	equ	[$-2].lw	;(Search-table count, set by Init).
	mov	di,offset (SrchT-@)
@TA1	equ	[$-2].lw	;(Search-table start, set by Init).
	push	cs
	pop	es
	repne	scasw		;"Scan" for index in search table.
	lea	bp,[di-2]	;Index is at [di-2], due to "scasw".
	call	CBGet		;Get cache entry from XMS to buffer.
	jc s	UnLnkE		;If any XMS error, go exit below.
;
; Subroutine to delete a cache index from the binary-search table.
;
Delete:	mov	di,bp		;Point DI-reg. to index in table.
	lea	si,[di+2]	;Set move-down source and count.
	mov	cx,[bx+STLmt-@]
	sub	cx,si
	shr	cx,1
	push	cs		;Set binary-search table segment and
	pop	es		;  move down remaining table indexes.
	rep	movs es:[di].lw,cs:[si].lw
	stosw			;Put deleted index on free-list.
	push	ds		;Reload this driver's ES-reg.
	pop	es
	sub	[bx+STLmt-@].lw,2  ;Decrement table limit.
;
; Subroutine to unlink a cache-table entry from the LRU list.
;
UnLink:	mov	cx,[bx+CBLst-@]	;Get current entry's "last" index.
	mov	dx,[bx+CBNxt-@]	;Get current entry's "next" index.
	cmp	cx,-1		;Is this entry top-of-list?
	je s	UnLnk1		;Yes, "promote" next entry.
	mov	ax,cx		;Get "prior" entry into work buffer.
	call	WBGet
	jc s	UnlnkE		;If XMS error, go exit below!
	mov	WBNxt,dx	;Update working entry's "next" index.
	inc	bx		;Put work-buffer entry back into XMS.
	call	WBPut
	jnc s	UnLnk2		;If no XMS error, check end-of-list.
UnLnkE:	pop	cx		;XMS error!  Discard return address.
	jmp	EndIO		;Go post XMS error code and get out!
UnLnk1:	mov	[bx+LUTop-@],dx	;Make next entry top-of-list.
UnLnk2:	cmp	dx,-1		;Is this entry end-of-list?
	jne s	UnLnk3		;No, link next to prior entry.
	mov	[bx+LUEnd-@],cx	;Make prior entry end-of-list.
	ret			;Exit.
UnLnk3:	mov	ax,dx		;Get "next" entry into work buffer.
	call	WBGet
	jc s	UnLnkE		;If XMS error, go exit above!
	mov	WBLst,cx	;Update working entry's "last" index.
	inc	bx		;Put work-buffer entry back into XMS.
	call	WBPut
	jc s	UnLnkE		;If any XMS error, go exit above!
	ret			;Exit.
HMALEN1	equ	($-CDReq)	;"Small driver" HMA logic length.
HMALEN2	equ	($-UdmaIO)	;"Large driver" HMA logic length.
;
; Cache Table Definitions.
;
SrchT	equ	$		;Start of binary-search table.
ResEnd	equ	SrchT+STACK	;End of "resident" logic and stack.
;
; Initialization Cache-Sizes Table.
;
; NOTE:  /Snnnn "large caches" of more than 250-MB should be used only
; on a Pentium IV CPU or better, due to all the required binary-search
; table handling.   At least 512-MB of upper memory must be available,
; and more is HIGHLY recommended, so other applications will also have
; a bit of XMS memory!
;
CachSiz	dw	TBLCT		;/S5 5-MB:  Fixed 640 cache blocks.
	db	16		;     8K granularity:  16 sectors per
	db	4		;        block, 4 sector-shift count.
	db	(1024/256)	;     1024 binary-search midpoint.
	db	8		;     8K cache-table XMS memory.
	db	"   5"		;     5-MB "title" message bytes.
	dw	(TBLCT*2)	;/S10 10-MB:  Fixed 1280 cache blocks.
	db	16,4		;     8K granularity values.
	db	(2048/256)	;     2048 binary-search midpoint.
	db	16		;     16K cache-table XMS memory.
	db	"  10"		;     10-MB "title" message bytes.
	dw	(TBLCT*2)	;/S20 20-MB:  Fixed 1280 cache blocks.
	db	32,5		;     16K granularity values.
	db	(2048/256)	;     2048 binary-search midpoint.
	db	16		;     16K cache-table XMS memory.
	db	"  20"		;     20-MB "title" message bytes.
	dw	(TBLCT*2)	;/S40 40-MB:  Fixed 1280 cache blocks.
	db	64,6		;     32K granularity values.
	db	(2048/256)	;     2048 binary-search midpoint.
	db	16		;     16K cache-table XMS memory.
	db	"  40"		;     40-MB "title" message bytes.
VTBlkCt	dw	(80*16)		;/Snnnn:  nnnn * 16 cache blocks.
	db	128,7		;     64K granularity values.
	db	(2048/256)	;     2048 to 16384 search midpoint.
	db	24		;     24K to 192K cache-table XMS.
	db	"  80"		;     nnnn-MB "title" message bytes.
;
; Initialization Variables.
;
	ifndef	UIDES		;For the normal UIDE driver --
HMASize	dw	HMALEN1		;  HMA memory needed (set by Init).
	endif
PCISubC	dw	00100h		;PCI subclass and function code.
XMHdl	dw	0		;XMS memory "handle" number.
XMAlloc	db	009h		;XMS V2.0 "allocate memory" command.
RFlag	db	0		;"Restricted memory" flag (no HMA).
;
; CD/DVD I-O Address Table.   The 1st word for each set is its DMA
;   controller base address, which is set by Init.   A value of -1
;   denotes an unused controller.   A value of 0 in the DVD1 table
;   indicates an old system having no DMA controller, whose CD/DVD
;   drives must run in "PIO mode" only.    The second word is each
;   controller's default "legacy" base I-O address.    For "native
;   PCI mode", all base I-O addresses are set during Init.   There
;   is an EXTRA address set, so that on old systems, the first set
;   of "legacy" addresses can be used even with no DMA controller.
;   Note that this table is used only during init.    At run-time,
;   a CD/DVD drive's addresses are taken from its parameter table.
;
DVD1Tbl	dw	0		;IDE1 primary-master parameters.
DVD1Pri	dw	NPDATA
	dw	0		;IDE1 secondary-master parameters.
DVD1Sec	dw	NSDATA
DVD2Tbl	dw	-1		;IDE2 primary-master parameters.
DVD2Pri	dw	APDATA
	dw	-1		;IDE2 secondary-master parameters.
DVD2Sec	dw	ASDATA
	dw	-1,0		;IDE3 primary-master parameters.
	dw	-1,0		;IDE3 secondary-master parameters.
	dw	-1,0		;IDE4 primary-master parameters.
	dw	-1,0		;IDE4 secondary-master parameters.
	dw	-1,0		;IDE5 primary-master parameters.
	dw	-1,0		;IDE5 secondary-master parameters.
	dw	-1,0		;IDE6 primary-master parameters.
	dw	-1,0		;IDE6 secondary-master parameters.
	dw	-1,0		;IDE7 primary-master parameters.
	dw	-1,0		;IDE7 secondary-master parameters.
	dw	-1,0		;IDE8 primary-master parameters.
	dw	-1,0		;IDE8 secondary-master parameters.
	dw	-1,0		;IDE9 primary-master parameters.
	dw	-1,0		;IDE9 secondary-master parameters.
DVDTEnd	equ	$		;End of CD/DVD addresses table.
DVDTSIZ	equ	DVD2Tbl-DVD1Tbl	;Size of a CD/DVD address table.
;
; Initialization UltraDMA "Mode" Table (digit count in low 4 bits).
;
Modes	dw	01602h		;Mode 0, ATA-16.
	dw	02502h		;Mode 1, ATA-25.
	dw	03302h		;Mode 2, ATA-33.
	dw	04402h		;Mode 3, ATA-44.
	dw	06602h		;Mode 4, ATA-66.
	dw	01003h		;Mode 5, ATA-100.
	dw	01333h		;Mode 6, ATA-133.
	dw	01663h		;Mode 7, ATA-166.
;
; Initialization Messages.
;
EDDBuff equ	$
TTLMsg	db	CR,LF,'UIDE'
	ifdef	UIDES
	db	'S'		;Include an "S" for UIDES.
	endif
	db	', 11-01-2007.   '
TTL2	db	'   5-MB Cache, $'
TTL3	db	'CD/DVD name is '
TTLName	db	'        .',CR,LF,'$'
IBMsg	db	'PCI test failed, BIOS I-O only!',CR,LF,'$'
NXMsg	db	'No XMS manager$'
DNMsg	db	' d'
CDName	db	'isk is '
DName	equ	$
DNEnd	equ	DName+40
BCMsg	db	'BAD$'
PCMsg	db	'IDE1'
PCMsg1	db	' Controller at I-O address '
PCMsg2	db	'    h, Chip I.D. '
PCMsg3	db	'        h.',CR,LF,'$'
NonUMsg	db	'Disks run by the BIOS:  '
NonUCt	db	'0. ',CR,LF,'$'
EDDMsg	db	'EDD BIOS$'
CHSMsg	db	'CHS'
UnitMsg	db	' data ERROR, D'
UnitNam	db	'rive '
UnitNo	db	'A: ',CR,LF,'$'
CDMsg	db	'CD'
CDMsgNo	db	'0:  '
CtlMsg	db	'IDE'
CtlrNo	db	'1 $'
PriMsg	db	'Primary-$'
SecMsg	db	'Secondary-$'
MstMsg	db	'master$'
SlvMsg	db	'slave$'
IEMsg	db	' Identify ERROR!',CR,LF,'$'
UEMsg	db	' is not UltraDMA!',CR,LF,'$'
VEMsg	db	'VDS init error$'
NDMsg	db	'Nothing to use$'
PRMsg	db	'No 80386+ CPU'
Suffix	db	'; UIDE'
	ifdef	UIDES
	db	'S'		;Include an "S" for UIDES.
	endif
	db	' not loaded!',CR,LF,'$'
;
; Driver Initialization Routine.
;
I_Init:	pushf			  ;Entry -- save CPU flags.
	push	ds		  ;Save all needed 16-bit registers.
	push	es
	push	ax
	push	bx
	push	dx
	push	cs		  ;Set this driver's DS-register.
	pop	ds
	les	bx,RqPkt	  ;Point to DOS "Init" packet.
	cmp	es:[bx].RPOp,0	  ;Is this really an "Init" packet?
	jne s	I_Quit		  ;No?  Reload regs. and exit quick!
	mov	ax,es:[bx].RPLen  ;Save "free" regular-memory limit.
	mov	@XEntry.lw,ax	  ;("Free" HMA is determined below).
	mov	ax,es:[bx].RPSeg
	mov	@XEntry.hw,ax
	mov	ax,cs		  ;Set "fixed" driver segment values.
	mov	@CDMain.hw,ax
	mov	@HDMain.hw,ax
	mov	@CDXSeg,ax
	mov	@RQLSeg,ax
	mov	@CCDSeg,ax
	mov	@RLRSeg,ax
	mov	@RLLSeg,ax
	mov	@RLMSeg,ax
	mov	@HMVSeg,ax
	mov	@DDLSeg,ax
	mov	@CRSSeg,ax
	mov	@HDXSeg,ax
	mov	@CEMSeg,ax
	mov	DvrTbl.hw,ax	;Set driver-address segment pointers.
	mov	Ud2Ptr.hw,ax
	mov	CDDPtr.hw,ax
	mov	es:[bx].RPStat,RPDON+RPERR  ;Set init packet defaults.
	mov	es:[bx].RPSeg,cs
	and	es:[bx].RPLen,0
	push	sp		;See if CPU is an 80286 or newer.
	pop	ax		;(80286+ push SP, then decrement it).
	cmp	ax,sp		;Did SP-reg. get saved "decremented"?
	jne s	I_Junk		;Yes, CPU is an 8086/80186, TOO OLD!
	pushf			;80386 test -- save CPU flags.
	push	07000h		;Try to set NT|IOPL status flags.
	popf
	pushf			;Get resulting CPU status flags.
	pop	ax
	popf			;Reload starting CPU flags.
	test	ah,070h		;Did any NT|IOPL bits get set?
	jnz s	I_386		;Yes, CPU is at least an 80386.
I_Junk:	mov	dx,(PRMsg-@)	;Display "No 80386+ CPU" message.
	call	I_Msg
I_Quit:	jmp	I_Exit		;Go reload registers and exit quick!
I_386:	pushad			;80386+ CPU:  Save 32-bit registers.
	les	si,es:[bx].RPCL	;Get command-line data pointer.
	xor	bx,bx		;Zero BX-reg. for relative logic.
I_NxtC:	mov	al,es:[si]	;Get next command-line byte.
	inc	si		;Bump pointer past first byte.
	cmp	al,0		;Is byte the command-line terminator?
	je s	I_TrmJ		;Yes, save current Int 13h vector.
	cmp	al,LF		;Is byte an ASCII line-feed?
	je s	I_TrmJ		;Yes, save current Int 13h vector.
	cmp	al,CR		;Is byte an ASCII carriage-return?
I_TrmJ:	je	I_Term		;Yes, save current Int 13h vector.
	cmp	al,'-'		;Is byte a dash?
	je s	I_NxtS		;Yes, see what next "switch" byte is.
	cmp	al,'/'		;Is byte a slash?
	jne s	I_NxtC		;No, check next command-line byte.
I_NxtS:	mov	ax,es:[si]	;Get next 2 command-line bytes.
	and	al,0DFh		;Mask out 1st byte's lower-case bit.
	cmp	al,'A'		;Is this byte an "A" or "a"?
	jne s	I_ChkB		   ;No, see if byte is "B" or "b".
	mov	al,(ASDATA-0100h)  ;Reverse all "Legacy IDE" addrs.
	mov	Ctl1Sec.lb,al
	mov	DVD1Sec.lb,al
	mov	al,(APDATA-0100h)
	mov	Ctl1Pri.lb,al
	mov	DVD1Pri.lb,al
	mov	al,(NSDATA-0100h)
	mov	Ctl2Sec.lb,al
	mov	DVD2Sec.lb,al
	mov	al,(NPDATA-0100h)
	mov	Ctl2Pri.lb,al
	mov	DVD2Pri.lb,al
	inc	si		;Point to next command-line byte.
I_ChkB:	cmp	al,'B'		;Is this byte a "B" or "b"?
	jne s	I_ChkL		;No, see if byte is "L" or "l".
	mov	ExEntrP,bx	;Reset "External Entry" routine ptr.
	mov	ax,(DvrEnd-@)	;Set stand-alone driver size.
	mov	VDSLn.lw,ax
	db	066h,0B8h	     ;Enable stand-alone disk logic
	db	(A20Er-GoMain),09Ah  ;  with a "call UdmaIO" at the
	dw	(UdmaIO-@)	     ;  label "GoMain" above.
	mov	[GoMain-1].dwd,eax
	mov	al,0EBh		     ;Disable all CD/DVD caching.
	mov	@NoCA1.lb,al
	db	0B8h		;Disable I-O through cache buffers.
	stc
	retf
	mov	UseCB1.lw,ax
	mov	RFlag,al	;Set "restricted" to exit below.
	mov	al,'$'		;Disable cache size in "title".
	mov	TTL2.lb,al
	inc	si		;Point to next command-line byte.
I_ChkL:	ifndef	UIDES		;For the normal UIDE driver --
	cmp	al,'L'		;  Is this byte an "L" or "l"?
	jne s	I_CkN1		;  No, see if switch bytes are "N1".
	mov	HMASize,HMALEN2	;  Force "large" upper-memory driver.
	inc	si		;  Point to next command-line byte.
	endif
I_CkN1:	cmp	ax,"1N"		;Are these 2 switch bytes "N1"?
	jne s	I_CkN2		;No, see if switch bytes are "N2".
	mov	BiosHD,bl	;Disable all hard-disk handling.
	inc	si		;Skip these 2 command-line bytes.
	inc	si
I_CkN2:	cmp	ax,"2N"		;Are these 2 switch bytes "N2"?
	jne s	I_ChkQ		;No, see if byte is "Q" or "q".
	mov	DVDTblB,ax	;Disable all CD/DVD drive handling.
	inc	si		;Skip these 2 command-line bytes.
	inc	si
I_ChkQ:	cmp	al,'Q'		;Is this byte a "Q" or "q"?
	jne s	I_ChkR		;No, see if byte is "R" or "r".
	mov	@DRQ.lb,075h	;Enable "data request" tests above.
	inc	si		;Point to next command-line byte.
I_ChkR:	cmp	al,'R'		;Is this byte an "R" or "r"?
	jne s	I_ChkS		;No, see if byte is "S" or "s".
	mov	RFlag,al	;Set "restricted memory" flag.
	inc	si		;Point to next command-line byte.
I_ChkS:	cmp	al,'S'		;Is this byte an "S" or "s"?
	jne s	I_ChkU		;No, see if byte is "U" or "u".
	call	GetSiz		;Get desired user cache size.
	jmp s	I_NxtJ		;Continue scanning for a terminator.
I_ChkU:	cmp	al,'U'		;Is this byte a "U" or "u"?
	jne s	I_ChkD		;No, see if 2 bytes are "D:" or "d:".
	inc	si		;Bump pointer past "U" or "u".
	mov	al,ah		;Get 2nd command-line byte.
	and	al,0DFh		;Mask out 1st lower-case bit (020h).
	cmp	al,'X'		;Is 2nd byte an "X" or "x"?
	jne s	I_NxtJ		;No, continue scan for a terminator.
	mov	NoDMA,al	;Set CD/DVD "No UltraDMA" flag.
	inc	si		;Bump pointer past "X" or "x".
I_NxtJ:	jmp	I_NxtC		;Go check next command-line byte.
I_ChkD:	cmp	ax,":D"		;Are next 2 bytes "D:" or "d:"?
	jne s	I_NxtJ		;No, continue scan for a terminator.
	inc	si		;Bump pointer past "device" switch.
	inc	si
	mov	di,(DvrNam-@)	;Blank out device name.
	mov	eax,"    "
	mov	[di],eax
	mov	[di+4],eax
I_Nam1:	mov	al,es:[si]	;Get next device-name byte.
	cmp	al,TAB		;Is byte a "tab"?
	je s	I_NxtJ		;Yes, handle above, "name" has ended!
	cmp	al,' '		;Is byte a space?
	je s	I_NxtJ		;Yes, handle above, "name" has ended!
	cmp	al,'/'		;Is byte a slash?
	je s	I_NxtJ		;Yes, handle above, "name" has ended!
	cmp	al,0		;Is byte the command-line terminator?
	je s	I_Term		;Yes, go test for UltraDMA controller.
	cmp	al,LF		;Is byte an ASCII line-feed?
	je s	I_Term		;Yes, go test for UltraDMA controller.
	cmp	al,CR		;Is byte an ASCII carriage-return?
	je s	I_Term		;Yes, go test for UltraDMA controller.
	cmp	al,'a'		;Ensure letters are upper-case.
	jb s	I_Nam2
	cmp	al,'z'
	ja s	I_Nam2
	and	al,0DFh
I_Nam2:	cmp	al,'!'		;Is this byte an exclamation point?
	je s	I_Nam3		;Yes, store it in device name.
	cmp	al,'#'		;Is byte below a pound-sign?
	jb s	I_Nam4		;Yes, Invalid!  Blank first byte.
	cmp	al,')'		;Is byte a right-parenthesis or less?
	jbe s	I_Nam3		;Yes, store it in device name.
	cmp	al,'-'		;Is byte a dash?
	je s	I_Nam3		;Yes, store it in device name.
	cmp	al,'0'		;Is byte below a zero?
	jb s	I_Nam4		;Yes, invalid!  Blank first byte.
	cmp	al,'9'		;Is byte a nine or less?
	jbe s	I_Nam3		;Yes, store it in device name.
	cmp	al,'@'		;Is byte below an "at sign"?
	jb s	I_Nam4		;Yes, invalid!  Blank first byte.
	cmp	al,'Z'		;Is byte a "Z" or less?
	jbe s	I_Nam3		;Yes, store it in device name.
	cmp	al,'^'		;Is byte below a carat?
	jb s	I_Nam4		;Yes, invalid!  Blank first byte.
	cmp	al,'~'		;Is byte above a tilde?
	ja s	I_Nam4		;Yes, invalid!  Blank first byte.
	cmp	al,'|'		;Is byte an "or" symbol?
	je s	I_Nam4		;Yes, invalid!  Blank first byte.
I_Nam3:	mov	[di],al		;Store next byte in device name.
	inc	si		;Bump command-line pointer.
	inc	di		;Bump device-name pointer.
	cmp	di,(DvrNam+8-@)	;Have we stored 8 device-name bytes?
	jb s	I_Nam1		;No, go get next name byte.
	jmp s	I_NxtJ		      ;Go get next command byte.
I_Nam4:	mov	[bx+DvrNam-@].lb,' '  ;Invalid!  Blank first byte.
	jmp s	I_NxtJ		      ;Go get next command byte.
I_Term:	cmp	[bx+DvrNam-@].lb,' '	    ;Is driver "name" valid?
	jne s	I_SetN			    ;Yes, drop any spaces.
	mov	[bx+DvrNam-@].dwd,"DVDU"    ;Set our default "name".
	mov	[bx+DvrNam+4-@].dwd,"   1"
I_SetN:	mov	si,(TTLName+8-@)	    ;Set "name" in title msg.
	mov	eax,[bx+DvrNam-@]
	mov	[si-8],eax
	mov	eax,[bx+DvrNam+4-@]
	mov	[si-4],eax
I_ScnN:	dec	si		;Decrement "name" pointer.
	cmp	[si].lb,' '	;Is this "name" byte a space?
	je s	I_ScnN			;Yes, scan for a non-space.
	mov	[si+1].dwd,0240A0D2Eh	;Set terminators after name.
	mov	ax,03513h		;Get and save Int 13h vector.
	call	I_In21
	push	es
	push	bx
	pop	eax
	mov	@Prv13A,eax
	mov	@Prv13B,eax
	cmp	GoMain.lb,09Ah	;Will this be the stand-alone driver?
	je	I_TTL		;Yes, go display driver "title" msg.
	ifdef	UIDES		;For the UIDES driver --
	cmp	VTBlkCt,200*16	;  200-MB cache or less desired?
	else			;For the normal UIDE driver --
	mov	ax,HMALEN2	;  Get "large" HMA logic size.
	cmp	ax,HMASize	;  "Large" upper-memory driver wanted?
	je s	I_LrgD		;  Yes, adjust driver size for it.
	cmp	VTBlkCt,200*16	;  200-MB cache or less desired?
	jbe s	I_GtHF		;  Yes, get HMA "free" memory.
	mov	HMASize,ax	;  Set "large driver" HMA size.
I_LrgD:	mov	ax,(NormSz2-@)	;  Set "large driver" memory size.
	mov	VDSLn.lw,ax
	cmp	VTBlkCt,255*16	;  255-MB cache or less desired?
	endif
	jbe s	I_GtHF		;Yes, get HMA "free" memory.
	inc	RFlag		;Greater may CRASH with V7.10 MS-DOS!
I_GtHF:	mov	ax,04A01h	;Get total HMA "free" memory.
	call	I_In2F
	movzx	cx,SFlag	;Get user's cache-size flag.
	cmp	ch,RFlag	;"Restricted" to normal memory?
	jne s	I_UprM		;Yes, set "upper-memory" driver size.
	ifdef	UIDES		;For the UIDES driver --
	sub	bx,HMALEN1	;  Enough memory for our HMA routines?
	else			;For the normal UIDE driver --
	sub	bx,HMASize	;  Enough memory for our HMA routines?
	endif
	jb s	I_NoHM		;No, we are stuck with normal memory!
	mov	dx,bx		;Set remaining "free" HMA in EDX-reg.
	cmp	dx,(TBLCT*2)	;Enough HMA for a small search table?
	jb s	I_NoHM		;No, we are stuck with normal memory!
	jcxz	I_Size		;If only 5-MB wanted, go set it up.
	cmp	dx,(TBLCT*4)	;Enough for our medium search table?
	jae s	I_Size		;Yes, put binary-search table in HMA.
I_NoHM:	inc	RFlag		;Set "restricted memory" flag.
I_UprM:	mov	ax,(ResEnd-@)	;Set our "normal memory" driver size.
	mov	VDSLn.lw,ax
	xor	edx,edx		;Get 32-bit "free" memory limit.
	xchg	dx,@XEntry.hw
	shl	edx,4
	add	edx,@XEntry
	xor	eax,eax		;Subtract 32-bit driver "base" size
	mov	ax,cs		;  for available search-table space.
	add	ax,((ResEnd-@)/16)
	shl	eax,4
	sub	edx,eax
	jb s	I_Min1		;Limit < driver?  Use minimum 5-MB!
	cmp	edx,65536	;64K or more of free memory?
	jb s	I_MinM		;No, see if we have minimum needed.
	mov	dx,0FFFFh	;Use 64K - 1 as free memory size.
I_MinM:	cmp	dx,(TBLCT*2)	;Enough for our minimum 5-MB cache?
	jae s	I_Size		;Yes, go see what cache is desired.
I_Min1:	mov	dx,(TBLCT*2)	;Our .SYS is 6K!  Use minimum 5-MB!
I_Size:	mov	si,(CachSiz-@)	;Point to our cache-sizes table.
	jcxz	I_Siz2		;If only 5-MB wanted, go set it up.
I_Siz1:	mov	ax,[si+10]	;Get next binary-search table size.
	shl	ax,1
	cmp	dx,ax		;Enough table space for next cache?
	jb s	I_Siz2		;No, default to current cache size.
	add	si,10		;Point to data for next cache size.
	dec	cx		;Did user request this cache size?
	jnz s	I_Siz1		;No, see if next cache can be used.
I_Siz2:	mov	ax,[si]		;Set cache-entry counts.
	add	@TC1,ax
	mov	@TC2,ax
	shl	ax,1		;Set binary-search limit addresses.
	add	@TE1,ax
	add	@TE2,ax
	cmp	ch,RFlag	;"Restricted" to normal memory?
	je s	I_Siz3		;No (using HMA), set "granularity".
	add	VDSLn.lw,ax	;Set final resident driver size.
I_Siz3:	mov	ax,[si+2]	;Set cache "granularity" values.
	mov	@GRAN1,al
	mov	@GRAN2,al
	mov	@GRAN3,al
	mov	@GRANSC,ah
	movzx	dx,al		;Save sectors-per-block for below.
	and	al,080h		;80-MB cache and up (64K granularity)
	or	XMAlloc,al	;  must issue V3.0 XMS "allocate" cmd.
	xor	ax,ax		;Set binary-search midpoint values.
	mov	ah,[si+4]
	mov	@MP1,ax
	add	@MP2,ax
	mov	ax,[si]		;Multiply number of cache blocks
	shr	dx,1		;  times (sectors-per-block / 2).
	mul	dx
	push	dx		;Get 32-bit cache data XMS size.
	push	ax
	pop	eax
	movzx	ecx,[si+5].lb	;Get cache-table XMS memory size.
	add	eax,ecx		;Save total required XMS memory.
	add	RqXMS,eax
	mov	eax,[si+6]	;Set cache size in "title" message.
	mov	TTL2.dwd,eax
I_TTL:	mov	dx,(TTLMsg-@)	;Display driver "title" message.
	call	I_Msg
	mov	dx,(TTL3-@)
	call	I_Msg
	mov	ax,VDSLn.lw	;Set initial driver stack pointers.
	mov	@Stack1,ax
	mov	@Stack2,ax
	xor	eax,eax		;Zero EAX-reg. for 20-bit addressing.
	mov	es,ax		;Point ES-reg. to low memory.
	mov	al,es:[HDISKS]	;Save our BIOS hard-disk count.
	and	BiosHD,al
	mov	ax,cs		;Set 20-bit VDS driver address.
	mov	VDSSg.lw,ax
	shl	eax,4
	mov	VDSAd,eax
	cli			      ;Avoid interrupts in VDS tests.
	test	es:[VDSFLAG].lb,020h  ;Are "VDS services" active?
	jz s	I_REnI		      ;No, re-enable CPU interrupts.
	mov	ax,08103h	;"Lock" this driver into memory.
	mov	dx,0000Ch
	call	I_VDS
	jc	I_ErrV		;If "lock" error, display msg. & exit!
	inc	VDSOf.lb	;Set initialization VDS "lock" flag.
I_REnI:	sti			;Re-enable CPU interrupts.
	mov	eax,VDSAd	;Get final driver 32-bit address.
	add	GDTPAdr,eax	;Relocate "real mode" GDT base addr.
	add	@WBAdr,eax	;Relocate "working" buffer address.
	add	@CBAdr,eax	;Relocate cache-entry buffer address.
	add	@CLAddr,eax	;Relocate command-list source address.
	mov	GDT_CSL,ax	;Set lower CS-descriptor address.
	shr	eax,16		;Set upper CS-descriptor address.
	mov	GDT_CSM,al
	mov	GDT_CSH,ah
	xor	edi,edi		;Get PCI BIOS "I.D." code.
	mov	al,001h
	call	I_In1A
	cmp	edx," ICP"	;Is PCI BIOS V2.0C or newer?
	je s	I_ScnC		;Yes, scan for all IDE controllers.
	mov	dx,(IBMsg-@)	;Display "No V2.0C+ PCI" message.
	call	I_Msg
	jmp s	I_XMgr		;Go see if we have an XMS manager.
I_ScnC:	mov	al,LBABuf.lb	;Get next "interface bit" value.
	and	ax,00003h
	or	ax,PCISubC	;"Or" in subclass & current function.
	call	I_PCIC		;Test for specific PCI class/subclass.
	rol	LBABuf.lb,4	;Swap both "interface bit" values.
	mov	al,LBABuf.lb	;Load next "interface bit" value.
	or	al,al		;Both "interface" values tested?
	jns s	I_ScnC		;No, loop back and test 2nd one.
	add	PCISubC.lb,004h	;More PCI function codes to try?
	jnc s	I_ScnC		;Yes, loop back & try next function.
	cmp	DVDTblA,(DVD1Tbl-@)  ;Any CD/DVD controllers found?
	ja s	I_Scn1		;Yes, see about "Native PCI" checks.
	add	DVDTblA,DVDTSIZ	;Save "legacy" addresses for oldies!
I_Scn1:	test	al,001h		;Have we tested "Native PCI" ctlrs.?
	mov	LBABuf.lb,093h	;(Set "Native PCI" interface bits).
	jz s	I_ScnC		;No, loop back and test them, also.
;
; These additional controller tests could be added, if necessary:
;
;	mov	ax,00400h	;Scan for "RAID" controller.
;	call	I_PCIC
;	mov	ax,00520h	;Scan for ADMA "Single-stepping" ctlr.
;	call	I_PCIC
;	mov	ax,00530h	;Scan for ADMA "Continuous DMA" ctlr.
;	call	I_PCIC
;	mov	ax,00600h	;Scan for standard SATA controller.
;	call	I_PCIC
;	mov	ax,00601h	;Scan for "AHCI" SATA controller.
;	call	I_PCIC
;
I_XMgr:	mov	ax,04300h	;Inquire if we have an XMS manager.
	call	I_In2F
	mov	dx,(NXMsg-@)	;Point to "No XMS manager" message.
	cmp	al,080h		;Is an XMS manager installed?
	jne s	I_ErrJ		;No?  Display error message and exit!
	mov	ax,04310h	;Get XMS manager "entry" address.
	call	I_In2F
	push	es		;Save XMS manager "entry" addresses.
	push	bx
	pop	@XEntry
	mov	ah,XMAlloc	;Request all necessary XMS memory.
	mov	edx,RqXMS
	call	I_XMS
	jz s	I_XLok		;If no error, "lock" our XMS memory.
I_XErr:	mov	dx,(VEMsg-@)	;Set up "XMS init error" message.
	mov	VEMsg.lw,"MX"
I_ErrJ:	jmp	I_Err		;Go display XMS error message & exit!
I_XLok:	mov	XMHdl,dx	;Save XMS buffer handle number.
	mov	ah,00Ch		;"Lock" our XMS memory.
	call	I_XMS
	jnz s	I_XErr		;If error, display message and exit!
	shl	edx,16		;Get unaligned 32-bit buffer address.
	or	dx,bx
	mov	eax,edx		;Copy 32-bit address to EAX-reg.
	jz s	I_XBAd		;Any low-order XMS buffer "offset"?
	mov	ax,0FFFFh	;Yes, align address to an even 64K.
	inc	eax
I_XBAd:	mov	@XBAddr,eax	;Save aligned "main buffer" address.
	add	@XBase,eax	;Initialize cache-data base address.
	mov	cx,ax		;Get buffer "offset" in XMS memory.
	sub	cx,dx
	mov	edx,000010000h	;Put command-list after XMS buffer.
	jcxz	I_PRDA		;Is buffer already on a 64K boundary?
	mov	edx,-32		;No, put command-list before buffer.
	dec	@XBase.hw	;Decrement cache-data base by 64K.
I_PRDA:	add	eax,edx		;Set our 32-bit PRD address.
	mov	PRDAd,eax
	mov	ax,[RqXMS+1].lw	;Get needed XMS in 256-KB increments.
	and	al,0FCh		;Reset lower 2 bits = Megabytes only.
	shl	eax,18		;Convert from Megabytes to bytes.
	add	eax,@XBase	;Add cache-data XMS base address.
	mov	@CTBase,eax	;Set cache-table XMS base address.
	cmp	GoMain.lb,09Ah	;Will this be the stand-alone driver?
	je s	I_HDCh		;Yes, forget about running diskettes!
	xor	ax,ax		;Point ES-reg. to low memory.
	mov	es,ax
	mov	al,es:[HDWRFL]	;Get BIOS "hardware installed" flag.
	test	al,001h		;Any diskettes on this system?
	jz s	I_HDCh		;No, scan for available hard-disks.
	mov	al,es:[MCHDWFL]	;Use diskette media-change bits
	and	al,011h		;  for our number of diskettes.
I_FScn:	test	al,001h		;Can next unit signal media-changes?
	jz s	I_FMor		;No?  CANNOT use this old "clunker"!
	push	ax		;Save our diskette counter.
	mov	al,0C0h		;Get "diskette" device-type flags.
	call	I_CHSD		;Get and save diskette's CHS values.
	pop	ax		;Reload our diskette counter.
	inc	@LastU		;Bump unit-table index.
I_FMor:	inc	HDUnit		;Bump BIOS unit number.
	inc	UnitNo.lb	;Bump error-message unit number.
	shr	al,4		;Another diskette to check?
	jnz s	I_FScn		;Yes, loop back and do next one.
I_HDCh:	cmp	BiosHD,0	;Any BIOS hard-disks to use?
	je	I_SCD		;No, scan for CD/DVD units to use.
	mov	HDUnit,080h	    ;Set 1st BIOS hard-disk unit.
	mov	UnitNam.dwd," ksi"  ;Set messages for hard-disks.
	mov	[UnitNo+1].lw,".h"
I_Next:	mov	ah,HDUnit	    ;Set unit no. in error message.
	mov	cx,2
	mov	si,(UnitNo-1-@)
	call	I_HexA
	mov	al,080h		;Get "BIOS disk" device-type.
	call	I_CHSD		;Get and save next disk's CHS values.
	mov	ah,041h		;Get EDD "extensions" for this disk.
	mov	bx,055AAh
	call	I_In13
	jc s	I_NoUJ		;If none, bump non-UltraDMA disk count.
	cmp	bx,0AA55h	;Did BIOS "reverse" our entry code?
	jne s	I_NoUJ		;No, bump non-UltraDMA disk count.
	test	cl,004h		;Does this disk have "EDD" extensions?
	jz s	I_NoUJ		;No, bump non-UltraDMA disk count.
	mov	si,(EDDBuff-@)	;Point to "EDD" input buffer.
	mov	[si].lw,30	;Set 30-byte buffer size.
	mov	ah,048h		;Get this disk's "EDD" parameters.
	call	I_In13
	jc s	I_ErED		;Error?  Display msg. & ignore!
	cmp	[si].lw,30	;Did we get at least 30 bytes?
	jb s	I_NoUJ		;No, bump non-UltraDMA disk count.
	cmp	[si+26].dwd,-1	;Valid "DPTE" pointer?
	je s	I_NoUJ		;No, bump non-UltraDMA disk count.
	les	si,[si+26]	;Get this disk's "DPTE" pointer.
	mov	bx,15		;Calculate "DPTE" checksum.
	xor	cx,cx
I_CkSm:	add	cl,es:[bx+si]
	dec	bx
	jns s	I_CkSm
	jcxz	I_EDOK		;If checksum O.K., use parameters.
I_ErED:	mov	dx,(EDDMsg-@)	;Display "EDD BIOS ERROR" message.
	call	I_Msg
	mov	dx,(UnitMsg-@)
	call	I_Msg
I_NoUJ:	jmp s	I_NoUD		;Go bump number of non-UltraDMA disks.
I_EDOK:	mov	bx,00010h	;Initialize IDE unit number index.
	and	bl,es:[si+4]
	shr	bl,4
	mov	ax,es:[si]	;Get disk's IDE base address.
	mov	CtlrNo.lb,'1'	;Reset display to "Ctlr. 1".
	mov	si,(Ctl1Tbl-@)	;Point to IDE address table.
I_ITbl:	cmp	[si].lw,-1	;Is this IDE table active?
	je s	I_NoUJ		;No, go bump non-UltraDMA disk count.
	cmp	ax,[si+2]	;Is disk on this primary channel?
	je s	I_ChMS		;Yes, set disk channel and unit.
	inc	bx		;Adjust index for secondary channel.
	inc	bx
	cmp	ax,[si+4]	;Is disk on this secondary channel?
	je s	I_ChMS		;Yes, set disk channel and unit.
	inc	bx		;Adjust values for next controller.
	inc	bx
	add	si,CTLTSIZ
	inc	CtlrNo.lb
	cmp	si,(CtlTEnd-@)	;More IDE controllers to check?
	jb s	I_ITbl		;Yes, loop back and check next one.
	jmp s	I_NoUJ		;Go bump non-UltraDMA disk count.
I_ChMS:	push	bx		;Save disk's caching unit number.
	mov	IdeDA,ax	;Save disk's base IDE address.
	add	ax,CDSEL	;Point to IDE device-select register.
	xchg	ax,dx
	mov	al,bl		;Get drive-select command byte.
	shl	al,4
	or	al,LBABITS
	out	dx,al		;Select master or slave disk.
	push	ax		;Save drive-select and caching unit.
	push	bx
	mov	dx,(CtlMsg-@)	;Display IDE controller number.
	call	I_Msg
	pop	ax		;Reload caching unit number.
	mov	dx,(PriMsg-@)	;Point to "Primary" channel name.
	test	al,002h		;Is this a primary-channel disk?
	jz s	I_PRNm		;Yes, display "Primary" channel.
	mov	dx,(SecMsg-@)	;Point to "Secondary" channel name.
I_PRNm:	call	I_Msg		;Display disk's IDE channel name.
	pop	ax		;Reload drive-select byte.
	mov	dx,(MstMsg-@)	;Point to "master" disk name.
	test	al,010h		;Is this disk the master?
	jz s	I_MSNm		;Yes, display "master" name.
	mov	dx,(SlvMsg-@)	;Point to "slave" disk name.
I_MSNm:	call	I_Msg		;Display disk's master/slave name.
	call	I_ValD		;Validate disk as an UltraDMA unit.
	pop	ax		;Reload caching unit number.
	jc s	I_UDEr		;If any errors, display message.
	mov	bx,@LastU	;Change this disk's device-type to
	mov	[bx+TypeF-@],al	;  "IDE" and include channel/unit.
	jmp s	I_More		;Go check for any more BIOS disks.
I_UDEr:	call	I_Msg		;NOT UltraDMA -- Display error msg.
I_NoUD:	cmp	GoMain.lb,09Ah	;Is this the "stand alone" driver?
	je s	I_IgnD		;Yes, ignore disk and scan for more.
	inc	NonUCt		;Bump number of non-UltraDMA disks.
	cmp	NonUCt,'9'	;Over 9 non-UltraDMA hard disks?
	jbe s	I_More		       ;No, check for more disks.
	mov	NonUCT.dwd,00D2E2B39h  ;Set 9+ non-UltraDMA count.
I_More:	inc	@LastU		       ;Bump our unit-table index.
I_IgnD:	inc	HDUnit		   ;Bump BIOS unit number.
	cmp	@LastU,MAXBIOS	   ;More entries in our units table?
	jae s	I_SCD		   ;No, check for any CD/DVD drives.
	dec	BiosHD		   ;More BIOS disks to check?
	jnz	I_Next		   ;Yes, loop back and do next one.
I_SCD:	xor	bx,bx		   ;Zero BX-reg. for CD/DVD logic.
	mov	DNMsg.lw," ,"	   ;Change start of disk "name" msg.
	mov	CtlrNo.lb,'0'	   ;Reset display controller number.
I_SCD1:	test	USNdx,001h	   ;Is this a new I-O channel?
	jnz s	I_SCD2		   ;No, test for new controller.
	add	DVDTblB,4	   ;Bump to next I-O addresses.
I_SCD2:	cmp	USNdx,004h	   ;Is this a new controller?
	jb s	I_SCD3		   ;No, get unit's I-O addresses.
	inc	CtlrNo.lb	   ;Bump display controller number.
	mov	USNdx,bl	   ;Reset unit-select index.
I_SCD3:	mov	si,DVDTblB	   ;Load I-O parameter pointer.
	cmp	si,(DVDTEnd-@)	   ;Any more IDE units to check?
	jae s	I_AnyN		   ;No, check for any drives to use.
	mov	al,USNdx	   ;Set indicated unit-select byte.
	shl	al,4
	or	al,MSEL
	mov	USelB,al
	inc	USNdx		   ;Bump unit-select index.
	mov	eax,[si]	   ;Get this unit's I-O addresses.
	cmp	ax,-1		   ;Any more controllers to check?
	je s	I_AnyN		   ;No, check for any drives to use.
	mov	[bx+DMAAd-@],eax   ;Set this unit's I-O addresses.
	cmp	NoDMA,bl	   ;Is all UltraDMA disabled?
	jne s	I_SCD4		   ;Yes, default to no DMA address.
	or	ax,ax		   ;"Legacy IDE" with no DMA ctlr.?
	jnz s	I_SCD5		   ;No, go validate this unit.
I_SCD4:	or	[bx+DMAAd-@].lw,-1 ;Invalidate UltraDMA address.
I_SCD5:	call	I_VCD		   ;Validate unit as ATAPI CD/DVD.
	jc s	I_SCD1		   ;If invalid, merely ignore it.
	mov	si,UTblP	   ;Update unit-table parameters.
	mov	eax,[bx+DMAAd-@]
	mov	[si],eax
	mov	al,USelB
	mov	[si+6],al
	add	si,20		   ;Update unit-table pointer.
	mov	UTblP,si
	inc	[bx+CDUnits-@].lb  ;Bump number of active units.
	inc	CDMsgNo		   ;Bump display unit number.
	cmp	si,(UTblEnd-@)	   ;Can we install another drive?
	jb	I_SCD1		   ;Yes, loop back & check for more.
I_AnyN:	cmp	NonUCt.lb,'0'	;Do we have any non-UltraDMA disks?
	je s	I_AnyD		;No, check for any drives to use.
	mov	dx,(NonUMsg-@)	;Display "Non-UltraDMA disks" msg.
	call	I_Msg
I_AnyD:	mov	dx,(NDMsg-@)	;Point to "Nothing to use" message.
	movzx	ax,CDUnits.lb	;Get total CD/DVD and BIOS drives.
	add	ax,@LastU	;Did we find any drives to use?
	jz	I_Err		;No?  Display error message and exit!
	shl	RFlag,1		;"Restricted" to normal memory?
	jnz	I_Done		;Yes, set DOS "init" packet results.
	mov	ax,04A02h	;Request needed memory in the HMA.
	mov	bx,@TC2
	shl	bx,1
	ifdef	UIDES
	add	bx,HMALEN1
	else
	add	bx,HMASize
	endif
	call	I_In2F
	mov	ax,es		;Get returned HMA segment address.
	cmp	ax,-1		;Is segment = 0FFFFh?
	jne s	I_HMA1		;No, VERY unusual but not an error.
	cmp	ax,di		;Is offset also = 0FFFFh?
	jne s	I_HMA1		;No, set binary-search table addr.
	mov	eax," AMH"	;BAAAD News!  Set up HMA error msg.
	jmp s	I_HMA2		;Go display error message and exit!
I_HMA1:	ifndef	UIDES		;For the normal UIDE driver --
	cmp	HMASize,HMALEN1	;  Using the "small" 1.75K driver?
	jne	I_HMA4		;  No, set up "large" driver below.
	endif
	push	ax		;Set CD/DVD entry routine pointer. 
	push	di
	pop	eax
	mov	@CDMain,eax
	add	ax,(RLRead-CDReq)  ;Set CD/DVD "RLRead" driver ptr.
	mov	CDDPtr,eax
	add	ax,(UdmaIO-RLRead) ;Set disk "UdmaIO" driver ptrs.
	mov	DvrTbl,eax
	mov	UD2Ptr,eax
	add	ax,(Cache-UdmaIO)  ;Set cache entry routine pointer.
	mov	@HDMain,eax
	lea	ax,[di-(CDReq-@)]  ;Get starting HMA logic offset.
	add	@DspTB1,ax	   ;Adjust CD/DVD table pointers.
	add	@DspTB2,ax
	add	@DspTB3,ax
	add	@DspTB4,ax
	add	@DspOfs,ax	;Adjust CD/DVD dispatch offset.
	ifdef	UIDES		;For the UIDES driver --
	add	di,HMALEN1	;  Set binary-search table address.
	mov	@TA1,di
	mov	ax,@TC2		;  Set binary-search table limits.
	shl	ax,1
	add	ax,di
	mov	@TE1,ax
	mov	@TE2,ax
	mov	ax,@MP1		;  Set binary-search midpoint addrs.
	dec	ax
	dec	ax
	add	ax,di
	mov	@MP2,ax
	mov	ah,005h		;  Issue "A20 local-enable" request.
	call	I_XMS
	else			;For the normal UIDE driver --
	call	I_HSet		;  Set HMA initialization parameters.
	endif
	jz s	I_HMA3		;If no "A20" error, move up logic.
	mov	eax," 02A"	;BAAAD News!  Set up A20 error msg.
I_HMA2:	mov	VEMsg.dwd,eax
	jmp	I_ErrV		;Go display error message & exit!
I_HMA3:	mov	si,(CDReq-@)	;Get HMA move addresses.
	les	di,@CDMain
	ifdef	UIDES		;For the UIDES driver --
	mov	cx,(HMALEN1/2)	;  Set length of our HMA logic.
	else			;For the normal UIDE driver --
	jmp s	I_HMA5		;  Go move small-driver logic to HMA.
I_HMA4:	push	ax		;  Large driver -- Set "UdmaIO" ptrs.
	push	di
	pop	eax
	mov	DvrTbl,eax
	mov	UD2Ptr,eax
	add	ax,(Cache-UdmaIO)  ;Set cache routine entry pointer.
	mov	@HDMain,eax
	call	I_HSet		;  Set HMA initialization parameters.
	jnz s	I_HMA2		;  If A20 error, display msg. & exit!
	mov	al,09Ah		;  Set disk driver subroutine opcodes.
	mov	[@UCBAd-1].lb,al
	mov	[@USAAd-1].lb,al
	mov	[@DDIAd-1].lb,al
	mov	[@DDRAd-1].lb,al
	push	cs		;  Set disk driver subroutine addrs.
	push	(UseCB-@)
	pop	eax
	mov	@UCBAd.dwd,eax
	mov	ax,(SetAdr-@)
	mov	@USAAd.dwd,eax
	mov	ax,(IniDMA-@)
	mov	@DDIAd.dwd,eax
	mov	ax,(RunDMA-@)
	mov	@DDRAd.dwd,eax
	mov	si,(UdmaIO-@)	;  Get HMA "move" addresses.
	les	di,DvrTbl
I_HMA5:	mov	cx,HMASize	;  Set length of our HMA logic.
	shr	cx,1
	endif
	rep	movsw		;Move all desired logic up to the HMA.
	mov	ah,006h		;Issue "A20 local-disable" request.
	call	I_XMS
I_Done:	xor	ax,ax		;Load & reset driver length.
	xchg	ax,VDSLn.lw
	les	bx,RqPkt	;Set result values in "init" packet.
	mov	es:[bx].RPLen,ax
	mov	es:[bx].RPStat,RPDON
	mov	ax,(DevInt-@)	;Set run-time "Device Interrupt" ptr.
	mov	DevIntP,ax
	push	cs		;Set ES-reg. same as DS-reg.
	pop	es
	jmp	I_End		;Go clear driver's stack and exit.
I_ErrV:	mov	dx,(VEMsg-@)	;Point to "VDS init error" message.
I_Err:	push	dx		;Init ERROR!  Save message pointer.
	shr	VDSOf.lb,1	;Was driver "locked" by VDS?
	jnc s	I_XDis		;No, check on XMS memory.
	mov	ax,08104h	;"Unlock" this driver from memory.
	xor	dx,dx
	call	I_VDS
I_XDis:	mov	dx,XMHdl	;Load our XMS memory "handle".
	or	dx,dx		;Did we reserve any XMS memory?
	jz s	I_LDMP		;No, reload pointer & display msg.
	mov	ah,00Dh		;Unlock and "free" our XMS memory.
	push	dx
	call	I_XMS
	mov	ah,00Ah
	pop	dx
	call	I_XMS
I_LDMP:	pop	dx		;Reload error message pointer.
	call	I_Msg		;Display desired error message.
	mov	dx,(Suffix-@)	;Display error-message suffix.
	call	I_Msg
	jmp	I_Fail		;Go reload all registers and exit.
;
; Subroutine to get the user's desired /Snnnn cache size.
;
GetSiz:	inc	si		;Bump scan pointer past "S" or "s".
	mov	di,(VTBlkCt-@)	;Point to variable cache-block count.
	mov	cl,(2048/256)	;Set default 80-MB cache parameters.
	mov	ch,24
	mov	[di+4],cx
	mov	[di+6].dwd,"    "  ;Reset variable "title" bytes.
GtSiz0:	mov	[di].lw,08000h	;Invalidate cache-block count.
GtSiz1:	mov	al,es:[si]	;Get next command-line byte.
	cmp	al,'9'		;Is byte greater than a '9'?
	ja s	GtSiz2		;Yes, go see what is in bucket.
	cmp	al,'0'		;Is byte less than a '0'?
	jb s	GtSiz2		;Yes, go see what is in bucket.
	inc	si		;Bump scan pointer past digit.
	cmp	[di+6].lb,' '	;Have we already found 4 digits?
	jne s	GtSiz0		;Yes, set INVALID & keep scanning.
	push	[di+7].dwd	;Shift "title" bytes 1 place left.
	pop	[di+6].dwd
	mov	[di+9],al	;Insert next "title" message byte.
	mov	cx,[di]		;Multiply current block size by 10.
	shl	[di].lw,2
	add	[di],cx
	shl	[di].lw,1
	and	ax,0000Fh	;"Add in" next cache size digit.
	add	[di],ax
	jmp s	GtSiz1		;Go scan for more cache-size digits.
GtSiz2:	mov	ax,[di]		;Get user's desired cache Megabytes.
	xor	cx,cx		;Set 5-MB cache size flag.
	cmp	ax,10		;Does user want only a 5-MB cache?
	jb s	GTSizX		;Yes, set cache-size flag and exit.
	inc	cx		;Set 10-MB cache size flag.
	cmp	ax,20		;Does user want the 10-MB cache?
	jb s	GTSizX		;Yes, set cache-size flag and exit.
	inc	cx		;Set 20-MB cache size flag.
	cmp	ax,40		;Does user want the 20-MB cache?
	jb s	GTSizX		;Yes, set cache-size flag and exit.
	inc	cx		;Set 40-MB cache size flag.
	cmp	ax,80		;Does user want the 40-MB cache?
	jb s	GTSizX		;Yes, set cache-size flag and exit.
	inc	cx		;Set "large" cache size flag.
	cmp	ax,1023		;Is user cache size invalid or > 1-GB?
	ja s	GtSizE		;Yes, ignore it & reset 80-MB default.
	cmp	ax,128		;Is cache size 80-MB to 127-MB?
	jb s	GTSiz3		;Yes, set user cache-block count.
	shl	[di+4].lw,1	;Double variable cache parameters.
	cmp	ax,256		;Is cache size 128-MB to 255-MB?
	jb s	GTSiz3		;Yes, set user cache-block count.
	shl	[di+4].lw,1	;Double variable cache parameters.
	cmp	ax,512		;Is cache size 256-MB to 511-MB?
	jb s	GTSiz3		;Yes, set user cache-block count.
	shl	[di+4].lw,1	;Double variable cache parameters.
GtSiz3:	shl	[di].lw,4	;Multiply blocks by 16 per Megabyte.
	jmp s	GtSizX		;Go set cache-size flag and exit.
GtSizE:	mov	[di],(TBLCT*2)	   ;Error!  Reset default 80-MB cache.
	mov	[di+6].dwd,"08  "
GtSizX:	mov	SFlag,cl	   ;Set desired cache-size flag.
	ret			   ;Exit.
;
; Subroutine to test for and set up a specific PCI disk controller.
;
I_PCIC:	mov	IdeDA,ax	     ;Save subclass & function codes.
	and	LBABuf.hw,0	     ;Reset PCI device index.
I_PCI1:	cmp	RqIndex,(CtlTEnd-@)  ;More space in address tables?
 	jae s	GtSizX		     ;No, go exit above.
	mov	ax,IdeDA	;Test PCI class 1, subclass/function.
	mov	ecx,000010003h	;(Returns bus/device/function in BX).
	xchg	ax,cx
	mov	si,LBABuf.hw
	call	I_In1A
	jc s	GtSizX		;Controller not found -- exit above.
	inc	LBABuf.hw	;Bump device index for another test.
	xor	di,di		;Get controller Vendor & Device I.D.
	call	I_PCID
	push	ecx		;Save Vendor and Device I.D.
	mov	di,32		;Save DMA controller base address.
	call	I_PCID
	xchg	ax,cx
	and	al,0FCh
	mov	DMAAd,ax
	mov	si,(PCMsg2-@)	;Set controller address in message.
	call	I_Hex
	mov	si,(PCMsg3-@)	;Set vendor I.D. in message.
	pop	ax
	call	I_Hex
	pop	ax		;Set Device I.D. in message.
	call	I_Hex
	mov	di,4		;Get low-order PCI command byte.
	call	I_PCID
	not	cl		;Get "Bus Master" & "I-O Space" bits.
	and	cl,005h		;Is controller using both BM and IOS?
	jz s	I_PCI2		;Yes, go save all controller data.
	mov	dx,(BCMsg-@)	;Display "BAD controller!", and hope
	call	I_Msg		;  our user can find a "better" BIOS!
	mov	dx,(PCMsg1-@)
	call	I_Msg
I_PCIJ:	jmp s	I_PCI1		;Go test for more same-class ctlrs.
I_PCI2:	mov	si,RqIndex	;Get current I-O address table ptr.
	mov	ax,DMAAd	;Set controller DMA base address.
	mov	[si],ax
	test	LBABuf.lb,001h	;Is this a "Native PCI" controller?
	jz s	I_PCI3		;No, set CD/DVD table addresses.
	mov	di,16		;Set primary-channel base address.
	call	I_PCID
	and	cl,0FCh
	mov	[si+2],cx
	mov	di,24		;Set secondary-channel base address.
	call	I_PCID
	and	cl,0FCh
	mov	[si+4],cx
I_PCI3:	cld			;Ensure FORWARD "string" commands.
	mov	di,DVDTblA	;Point to CD/DVD addresses table.
	push	cs
	pop	es
	lodsd			;Set CD/DVD primary addresses.
	stosd
	add	ax,8		;Set CD/DVD secondary addresses.
	stosw
	movsw
	mov	DVDTblA,di	;Bump CD/DVD table pointer.
	mov	dx,(PCMsg-@)	;Display all this controller's data.
	call	I_Msg
	inc	[PCMsg+3].lb	;Bump controller number in message.
	add	RqIndex,CTLTSIZ	;Bump controller address-table ptr.
	jmp s	I_PCIJ		;Go test for more same-class ctlrs.
;
; Subroutine to "validate" an UltraDMA hard-disk.
;
I_ValD:	xor	bx,bx		;Point ES-reg. to low memory.
	mov	es,bx
	mov	dx,[bx+IdeDA-@]	;Issue "Identify Device" command.
	add	dx,CCMD
	mov	al,0ECh
	out	dx,al
	mov	bx,BIOSTMR	;Point to low-memory BIOS timer.
	mov	cl,RDYTO	;Set I-O timeout limit in CL-reg.
	add	cl,es:[bx]
I_IDly:	cmp	cl,es:[bx]	;Has our command timed out?
	je s	I_DErr		;Yes, set "Identify" message & exit.
	in	al,dx		;Get IDE controller status.
	test	al,BSY+RDY	;Controller or disk still busy?
	jle s 	I_IDly		;Yes, loop back and check again.
	test	al,ERR		;Did command cause any errors?
	jz s	I_PIO		;No, read I.D. data using PIO mode.
I_DErr:	mov	dx,(IEMsg-@)	;Point to "Identify error" message.
I_SErr:	stc			;Set carry flag (error!) and exit.
	ret
I_PIO:	add	dx,(CDATA-CCMD)	;Point to PIO data register.
	in	ax,dx		;Read I.D. bytes 0 and 1.
	shl	ax,1		;Save "ATA/ATAPI" flag in carry bit.
	mov	cx,27		;Skip I.D. bytes 2-53 and
I_Skp1:	in	ax,dx		;  read I.D. bytes 54-55.
	loop	I_Skp1
	push	cs		;Point to disk-name message.
	pop	es
	mov	di,(DName-@)
	mov	cl,26		;Read & swap disk name into message.
I_RdNm:	xchg	ah,al		;(I.D. bytes 54-93.  Bytes 94-105 are
	stosw			;  also read but are ignored.   Bytes
	in	ax,dx		;  106-107 are left in the AX-reg.).
	loop	I_RdNm
	xchg	ax,bx		;Save "UltraDMA valid" flag in BL-reg.
	mov	cl,35		;Skip I.D. bytes 108-175 &
I_Skp2:	in	ax,dx		;  read I.D. bytes 176-177.
	loop	I_Skp2
	mov	bh,ah		;Save "UltraDMA mode" flags in BH-reg.
	mov	cl,167		;Skip remaining I.D. data.
I_Skp3:	in	ax,dx
	loop	I_Skp3
	mov	dx,(UEMsg-@)	;Point to "is not UltraDMA" message.
	rcr	bl,1		;Shift "ATA/ATAPI" flag into BL-reg.
	and	bl,082h		;ATAPI disk, or UltraDMA bits invalid?
	jle s	I_SErr		;Yes?  Exit & display error message.
	mov	di,(Modes-@)	;Point to UltraDMA mode table.
	or	bh,bh		;Will disk do UltraDMA mode 0?
	jz s	I_SErr		;No?  Exit & display message!
I_NxtM:	cmp	bh,bl		;Will disk do next UltraDMA mode?
	jb s	I_GotM		;No, use current mode.
	inc	di		;Point to next mode table value.
	inc	di
	shl	bl,1		;More UltraDMA modes to check?
	jnz s	I_NxtM		;Yes, loop back.
I_GotM:	push	[di].lw		;Save disk mode value.
	mov	si,(DNEnd-@)	;Point to end of disk name.
I_NxtN:	cmp	si,(DName-@)	;Are we at the disk-name start?
	je s	I_Name		;Yes, disk name is all spaces!
	dec	si		;Decrement disk name pointer.
	cmp	[si].lb,' '	;Is this name byte a space?
	je s	I_NxtN		;No, continue scan for non-space.
	inc	si		;Skip non-space character.
	mov	[si].lw," ,"	;End disk name with comma & space.
	inc	si
	inc	si
I_Name:	mov	[si].dwd,"-ATA"	;Set "ATA-" after disk name.
	add	si,4
	pop	ax		;Get disk "mode" value.
	jmp	I_VC17		;Go display disk name/"mode" & exit.
;
; Subroutine to set the BIOS CHS data for a hard-disk or diskette.
;
I_CHSD:	push	ax		;Save unit's device-type flag.
	mov	ah,008h		;Get BIOS CHS data for this unit.
	call	I_In13
	jc s	I_CHSE		;If BIOS error, zero both values.
	and	cl,03Fh		;Get sectors/head value (low 6 bits).
	jz s	I_CHSE		;If zero, ERROR!  Zero both values.
	inc	dh		;Get heads/cylinder (BIOS value + 1).
	jnz s	I_CHST		;If non-zero, save data in our table.
I_CHSE:	xor	cl,cl		;Error!  Zero CHS data for this unit.
	xor	dh,dh
I_CHST:	mov	di,@LastU	;Point to "active units" table.
	mov	al,HDUnit	;Set BIOS unit number in our table.
	mov	[di+Units-@],al
	mov	[di+CHSec-@],cl ;Set unit's CHS data in our table.
	mov	[di+CHSHd-@],dh
	pop	ax		;Reload and set device-type flag.
	mov	[di+TypeF-@],al
	or	cl,cl		;Valid CHS values for this unit?
	jnz s	I_VC7		;Yes, go exit below.
	mov	dx,(CHSMsg-@)	;Display "CHS data error" and exit.
	jmp	I_Msg
;
; Subroutine to "validate" an IDE unit as an ATAPI CD/DVD drive.
;
I_VCD:	mov	dx,[bx+DMAAd-@]	;Get unit UltraDMA command address.
	test	dl,001h		;Will this unit be using UltraDMA?
	jnz s	I_VC0		;No, just select "master" or "slave".
	in	al,dx		;Ensure any previous DMA is stopped!
	and	al,0FEh
	out	dx,al
I_VC0:	mov	dx,[bx+IdeDA-@]	;Point to IDE device-select register.
	add	dx,6
	mov	al,USelB	;Select IDE "master" or "slave" unit.
	out	dx,al
	call	ChkTO		;Await controller-ready.
	jc s	I_VC7		;If timeout, go exit below.
	mov	al,0A1h		;Issue "Identify Packet Device" cmd.
	out	dx,al
	call	ChkTO		;Await controller-ready.
	jc s	I_VC7		;If timeout, go exit below.
	test	al,DRQ		;Did we also get a data-request?
	jz s	I_VC6		;No, go set carry & exit below.
	sub	dx,7		;Point back to IDE data register.
	in	ax,dx		;Read I.D. word 0, main device flags.
	and	ax,0DF03h	;Mask off flags for an ATAPI CD/DVD.
	xchg	ax,si		;Save main device flags in SI-reg.
	mov	cx,26		;Skip I.D. words 1-26 (unimportant).
I_VC1:	in	ax,dx
	loop	I_VC1
	mov	di,(CDName-@)	;Point to drive "name" input buffer.
	push	cs
	pop	es
	mov	cl,20		;Read & swap words 27-46 into buffer.
I_VC2:	in	ax,dx		;(Manufacturer "name" of this drive).
	xchg	ah,al
	stosw
	loop	I_VC2
	mov	cl,7		;Skip I.D. words 47-52 (unimportant)
I_VC3:	in	ax,dx		;  and read I.D. word 53 into AX-reg.
	loop	I_VC3
	mov	UFlag,al	;Save UltraDMA "valid" flags.
	mov	cl,35		;Skip I.D. words 54-87 (unimportant)
I_VC4:	in	ax,dx		;  and read I.D. word 88 into AX-reg.
	loop	I_VC4
	mov	UMode.lb,ah	;Save posted UltraDMA "mode" value.
	mov	cl,167		;Skip all remaining I.D. data.
I_VC5:	in	ax,dx
	loop	I_VC5
	cmp	si,08500h	;Do device flags say "ATAPI CD/DVD"?
	je s	I_VC8		;Yes, see about UltraDMA use.
I_VC6:	stc			;Set carry flag on (error!).
I_VC7:	ret			;Exit.
I_VC8:	test	UFlag,004h	;Valid UltraDMA "mode" bits?
	jz s	I_VC9		;No, disable drive UltraDMA.
	cmp	UMode.lb,bl	      ;Can drive do mode 0 minimum?
	jne s	I_VC10		      ;Yes, display "Unit n:" msg.
I_VC9:	or	[bx+DMAAd-@].lb,1     ;Disable this drive's UltraDMA.
I_VC10:	mov	dx,(CDMsg-@)	      ;Display "CDn:  " message.
	call	I_Msg
	mov	dx,(PriMsg-@)	      ;Point to "Primary" message.
	test	[bx+IdeDA-@].lb,080h  ;Primary-channel drive?
	jnz s	I_VC11		      ;Yes, display "Primary" msg.
	mov	dx,(SecMsg-@)	   ;Point to "Secondary" message.
I_VC11:	call	I_Msg		   ;Display CD/DVD's IDE channel.
	mov	dx,(MstMsg-@)	   ;Point to "Master" message.
	cmp	USelB,SSEL	   ;Is this drive a "slave"?
	jnz s	I_VC12		   ;No, display "Master".
	mov	dx,(SlvMsg-@)	   ;Point to "Slave" message.
I_VC12:	call	I_Msg		   ;Display "Master" or "Slave".
	mov	si,(CDName+40-@)   ;Point to end of CD/DVD vendor name.
I_VC13:	cmp	si,(CDName-@)	   ;Are we at the vendor-name start?
	je s	I_VC14		   ;Yes, vendor name is all spaces!
	dec	si		   ;Decrement vendor-name pointer.
	cmp	[si].lb,' '	   ;Is this name byte a space?
	je s	I_VC13		   ;No, continue scan for non-space.
	inc	si		   ;Skip non-space character.
	mov	[si].lw," ,"	   ;End disk name with comma/space.
	inc	si		   ;Skip comma and space.
	inc	si
I_VC14:	test	[bx+DMAAd-@].lb,1  ;Will this drive use "PIO mode"?
	jz s	I_VC15		   ;No, set drive speed after name.
	mov	[si].dwd,"OIP"	   ;Set "PIO" after drive name.
	add	si,3
	jmp s	I_VC18		;Go set message terminators.
I_VC15:	mov	[si].dwd,"-ATA"	;Set "ATA-" after drive name.
	add	si,4
	mov	cx,UMode	;Initialize UltraDMA "mode" scan.
	mov	di,(Modes-2-@)
I_VC16:	inc	di		;Advance to next UltraDMA "mode".
	inc	di
	shr	cx,1		;Will drive do next "mode"?
	jnz s	I_VC16		;Yes, keep scanning for maximum.
	mov	ax,[di]		;Set UltraDMA "mode" in message.
I_VC17:	mov	cl,00Fh
	and	cl,al
	call	I_HexA
I_VC18:	mov	[si].dwd,0240A0D2Eh  ;Set message terminators.
	mov	dx,(DNMsg-@)	     ;Display vendor name & "mode".
	call	I_Msg
	clc			;Reset carry (no errors) and exit.
	ret
;
; Subroutine to initialize common parameters for use in the HMA.
;
	ifndef	UIDES		;For the normal UIDE driver --
I_HSet:	add	di,HMASize	;  Set binary-search table address.
	mov	@TA1,di
	mov	ax,@TC2		;  Set binary-search table limits.
	shl	ax,1
	add	ax,di
	mov	@TE1,ax
	mov	@TE2,ax
	mov	ax,@MP1		;  Set binary-search midpoint addrs.
	dec	ax
	dec	ax
	add	ax,di
	mov	@MP2,ax
	mov	ah,005h		;  Fall-through to issue "A20 local
				;    enable" request, and then exit.
	endif
;
; Subroutines to issue initialization "external" calls.
;
I_XMS:	call	@XEntry		;Issue desired XMS request.
	dec	ax		;Zero AX-reg. if success, -1 if error.
	jmp s	I_IntX		;Restore driver settings, then exit.
I_VDS:	push	bx		;VDS -- save our BX-reg.
	mov	di,(VDSLn-@)	;Point to VDS parameter block.
	push	cs
	pop	es
	int	04Bh		;Execute VDS "lock" or "unlock".
	pop	bx		;Reload our BX-register.
	jmp s	I_IntX		;Restore driver settings, then exit.
I_In13:	mov	dl,HDUnit	;Set BIOS unit in DL-reg.
	int	013h		;Issue BIOS data interrupt.
	jmp s	I_IntX		;Restore driver settings, then exit.
I_PCID:	push	bx		;Save PCI bus/device/function codes.
	push	si		;Save IDE address-table pointer.
	mov	al,00Ah		;Set "PCI doubleword" request code.
	call	I_In1A		;Get desired 32-bit word from PCI.
	pop	si		;Reload IDE address-table pointer.
	pop	bx		;Reload PCI bus/device/function.
	ret			;Exit.
I_In1A:	mov	ah,0B1h		;Issue PCI BIOS interrupt.
	int	01Ah
	jmp s	I_IntX		;Restore driver settings, then exit.
I_Msg:	mov	ah,009h		;Get DOS "display string" opcode.
I_In21:	int	021h		;Issue desired DOS request.
	jmp s	I_IntX		;Restore driver settings, then exit.
I_In2F:	int	02Fh		;Issue XMS interrupt.
I_IntX:	sti			;RESTORE all critical driver settings!
	cld			;(Never-NEVER "trust" external code!).
	push	cs
	pop	ds
	ret			;Exit.
;
; Subroutine to convert a 4-digit hex number to ASCII for messages.
;   At entry, the number is in the AX-reg., and the message pointer
;   is in the SI-reg.   At exit, the SI-reg. is updated and the CX-
;   reg. is zero.
;
I_Hex:	mov	cx,4		;Set 4-digit count.
I_HexA:	rol	ax,4		;Get next hex digit in low-order.
	push	ax		;Save remaining digits.
	and	al,00Fh		;Mask off next hex digit.
	cmp	al,009h		;Is digit 0-9?
	jbe s	I_HexB		;Yes, convert to ASCII.
	add	al,007h		;Add A-F offset.
I_HexB:	add	al,030h		;Convert digit to ASCII.
	mov	[si],al		;Store next digit in message.
	inc	si		;Bump message pointer.
	pop	ax		;Reload remaining digits.
	loop	I_HexA		;If more digits to go, loop back.
	ret			;Exit.
CODE	ends
	end
